###### Experiment Optimization Visualizer #####
#                                             #
#               By: Colin Yancey              #
#             Created March 2026              #
#            Last Edited: 03/04/2026          #
#                  Schulman Lab               #
#            Johns Hopkins University         #
#                                             #
###############################################

from pathlib import Path
from typing import Optional
import numpy as np
import jax.numpy as jnp
import time

import imageio.v2 as iio
import os

import matplotlib.pyplot as plt
from matplotlib import colors as mcolors

from ..results.optimizerResults import CompiledOptimizationData
from ..experiments.manager import OptimizeExperimentsManager
from ..utils.constants import UNIT_WEIGHTS
from ..utils.base import _getMinSecStr
from ..utils.helpers import _checkManagerType
from .plotting import *

###############################################


def saveExperimentResultsVisualization(
    compiledOptDataObject: CompiledOptimizationData,
    experimentSignalWhitelist: Optional[list[int | tuple[int, int] | tuple[int, int, int]]] = None,
    iterationList: Optional[list[int]] = None,
    fileLocation: Path = Path('./'),
    fileName: str = 'experimentResults',
    fps: int = 1,
    stackImages: bool = False,
    keepIndividualImages: bool = False,
    startLineColor: str = "#202020",
    endLineColor: str = "#30E0F0",
    dataPointColor: str = "#A0A0A0",
    dataPointAlpha: float = 0.5,
    dataPointSize: float = 14,
    timeUnits: str = "sec"
):
    _checkManagerType(compiledOptDataObject.ManagerObject, OptimizeExperimentsManager)

    if timeUnits not in UNIT_WEIGHTS:
        raise ValueError(f"Unknown timeUnits '{timeUnits}'. Available units: {list(UNIT_WEIGHTS.keys())}")

    optRes = compiledOptDataObject.OptimizerRunResultObject
    simRunner = compiledOptDataObject.SimulationRunnerObject
    optFrameworkObj = optRes.CRNOptimizationFrameworkObject
    exptGroups = compiledOptDataObject.ManagerObject.ExperimentGroups

    totalIter = optRes.LastIteration
    if iterationList is None:
        iterationList = np.arange(totalIter)
    else:
        iterationList = np.array(iterationList) - 1

    if len(iterationList) == 0:
        raise ValueError("iterationList must contain at least one iteration.")
    if np.any(iterationList < 0) or np.any(iterationList >= totalIter):
        raise ValueError(f"All entries in iterationList must be between 1 and {totalIter}, inclusive.")

    panelInfo = []

    def _add_all_signals(groupIdx: int, exptIdx: int):
        exptObj = exptGroups[groupIdx].Experiments[exptIdx]
        for signalIdx in range(len(exptObj.Signals)):
            candidate = (groupIdx, exptIdx, signalIdx)
            if candidate not in panelInfo:
                panelInfo.append(candidate)

    def _add_all_experiments(groupIdx: int):
        for exptIdx in range(len(exptGroups[groupIdx].Experiments)):
            _add_all_signals(groupIdx, exptIdx)

    if experimentSignalWhitelist is None:
        for groupIdx in range(len(exptGroups)):
            _add_all_experiments(groupIdx)
    elif len(experimentSignalWhitelist) == 0:
        raise ValueError(
            "If providing experimentSignalWhitelist, include at least one entry selecting a group, experiment, or signal."
        )
    else:
        for selection in experimentSignalWhitelist:
            if isinstance(selection, int):
                groupIdx = selection - 1
                if groupIdx < 0 or groupIdx >= len(exptGroups):
                    raise ValueError(f"Experiment Group {selection} does not exist.")
                _add_all_experiments(groupIdx)
                continue

            if len(selection) == 2:
                groupIdx = selection[0] - 1
                exptIdx = selection[1] - 1

                if groupIdx < 0 or groupIdx >= len(exptGroups):
                    raise ValueError(f"Experiment Group {selection[0]} does not exist.")
                if exptIdx < 0 or exptIdx >= len(exptGroups[groupIdx].Experiments):
                    raise ValueError(f"Experiment {selection[1]} is not contained within Experiment Group {selection[0]}.")

                _add_all_signals(groupIdx, exptIdx)
                continue

            if len(selection) == 3:
                groupIdx = selection[0] - 1
                exptIdx = selection[1] - 1
                signalIdx = selection[2] - 1

                if groupIdx < 0 or groupIdx >= len(exptGroups):
                    raise ValueError(f"Experiment Group {selection[0]} does not exist.")
                if exptIdx < 0 or exptIdx >= len(exptGroups[groupIdx].Experiments):
                    raise ValueError(f"Experiment {selection[1]} is not contained within Experiment Group {selection[0]}.")
                if signalIdx < 0 or signalIdx >= len(exptGroups[groupIdx].Experiments[exptIdx].Signals):
                    raise ValueError(
                        f"Signal {selection[2]} is not contained within Experiment {selection[1]} of Experiment Group {selection[0]}."
                    )

                candidate = (groupIdx, exptIdx, signalIdx)
                if candidate not in panelInfo:
                    panelInfo.append(candidate)
                continue

            raise ValueError(
                "Each whitelist entry must be either an int (group), a tuple of length 2 (group, experiment), or a tuple of length 3 (group, experiment, signal)."
            )

    if len(panelInfo) == 0:
        raise ValueError("No panels were selected for plotting.")

    uniqueExpts = sorted({(groupIdx, exptIdx) for groupIdx, exptIdx, _ in panelInfo})
    exptsByGroup = {}
    for groupIdx, exptIdx in uniqueExpts:
        if groupIdx not in exptsByGroup:
            exptsByGroup[groupIdx] = []
        exptsByGroup[groupIdx].append(exptIdx)

    parsedTimeCourseData = {}
    evalPts = {}
    for groupIdx, exptIdx in uniqueExpts:
        groupObj = exptGroups[groupIdx]
        exptObj = groupObj.Experiments[exptIdx]
        parsedTimeCourseData[(groupIdx, exptIdx)] = groupObj.TimePointsHandler.parseEvalPts(exptObj.TimeCourse)
        evalPts[(groupIdx, exptIdx)] = np.array(groupObj.TimePointsHandler.TotalEvalPts) / UNIT_WEIGHTS[timeUnits]

    print(f"Computing integration solutions...\n{"#"*40}")

    initCondCompsByGroup = {groupIdx: [] for groupIdx in exptsByGroup.keys()}
    exptSolutions = {(groupIdx, exptIdx): [] for groupIdx, exptIdx in uniqueExpts}

    startTime = time.time()

    firstIteration = iterationList[0]
    curLnParams = jnp.log(optRes.ParameterHistory[firstIteration])
    for groupIdx, exptIndices in exptsByGroup.items():
        groupObj = exptGroups[groupIdx]
        initCondInfo = optFrameworkObj.getStartingConditions(curLnParams, groupObj.CRNInfo)
        initCondCompsByGroup[groupIdx].append(initCondInfo.ComponentsDictionary)

        for exptIdx in exptIndices:
            exptSolutions[(groupIdx, exptIdx)].append(
                simRunner.runTimeCourse(
                    parsedTimeCourseData[(groupIdx, exptIdx)],
                    groupObj.CRNInfo,
                    initCondInfo
                )
            )

    iterTime = time.time() - startTime
    completionTimeEstimate = iterTime * (len(iterationList) - 1)
    print(
        "Integration solutions for the first selected iteration were found.\n"
        f"  Time: {iterTime:.3f} sec\n"
        f"  Estimated time to completion: {_getMinSecStr(completionTimeEstimate)}"
    )
    if completionTimeEstimate > 120:
        print("  If this is too long, it is recommended to use a sparser iteration sampling.")

    for iteration in iterationList[1:]:
        curLnParams = jnp.log(optRes.ParameterHistory[iteration])

        for groupIdx, exptIndices in exptsByGroup.items():
            groupObj = exptGroups[groupIdx]
            initCondInfo = optFrameworkObj.getStartingConditions(curLnParams, groupObj.CRNInfo)
            initCondCompsByGroup[groupIdx].append(initCondInfo.ComponentsDictionary)

            for exptIdx in exptIndices:
                exptSolutions[(groupIdx, exptIdx)].append(
                    simRunner.runTimeCourse(
                        parsedTimeCourseData[(groupIdx, exptIdx)],
                        groupObj.CRNInfo,
                        initCondInfo
                    )
                )

    print(
        "All integration solutions found.\n"
        f"  Total Time (including first iteration): {_getMinSecStr(time.time() - startTime)}\n"
        "Creating plots..."
    )

    panelYValues = [[None] * len(iterationList) for _ in range(len(panelInfo))]

    for panelIdx, (groupIdx, exptIdx, signalIdx) in enumerate(panelInfo):
        groupObj = exptGroups[groupIdx]
        exptObj = groupObj.Experiments[exptIdx]
        signalObj = exptObj.Signals[signalIdx]
        relevantIndices = signalObj.getRelevantComponentIndices(groupObj.CRNInfo.CompNameList)

        for iterIdx in range(len(iterationList)):
            relevantResults = exptSolutions[(groupIdx, exptIdx)][iterIdx][:, relevantIndices]
            if len(relevantIndices) == 1:
                rawSignalVals = relevantResults.reshape(-1)
            else:
                rawSignalVals = jnp.sum(relevantResults, axis=1)

            curVals = signalObj.getNormalizeVMAP(initCondCompsByGroup[groupIdx][iterIdx])(rawSignalVals)
            panelYValues[panelIdx][iterIdx] = curVals

    workspaceName = fileLocation / 'mp4 Output'
    os.makedirs(workspaceName, exist_ok=True)
    framePaths = []

    panelNum = len(panelInfo)
    squareGridDim = int(np.ceil(np.sqrt(panelNum)))
    gridLen = 4 * squareGridDim

    fig, axes = plt.subplots(squareGridDim, squareGridDim, figsize=(gridLen, gridLen))
    if squareGridDim == 1:
        axes = np.array([[axes]])
    flatAxes = axes.ravel()

    for m in range(panelNum, squareGridDim**2):
        flatAxes[m].axis('off')

    # For each column, find the lowest row that still contains a visible panel.
    lowestVisibleRowByColumn = {}
    for columnIndex in range(squareGridDim):
        visibleRows = [
            rowIndex
            for rowIndex in range(squareGridDim)
            if (rowIndex * squareGridDim + columnIndex) < panelNum
        ]
        if len(visibleRows) > 0:
            lowestVisibleRowByColumn[columnIndex] = max(visibleRows)

    for panelIdx, (groupIdx, exptIdx, signalIdx) in enumerate(panelInfo):
        groupObj = exptGroups[groupIdx]
        exptObj = groupObj.Experiments[exptIdx]
        signalObj = exptObj.Signals[signalIdx]

        ax = flatAxes[panelIdx]
        x = evalPts[(groupIdx, exptIdx)]

        ax.scatter(
            x,
            signalObj.DataPoints,
            s=dataPointSize,
            color=dataPointColor,
            alpha=dataPointAlpha,
            zorder=2,
            label='Data'
        )

        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)
        ax.grid(True, linestyle='--', alpha=0.3)
        ax.set_xlim(x[0], x[-1])
        ax.set_ylim(-0.03, 1.03)
        rowIndex = panelIdx // squareGridDim
        columnIndex = panelIdx % squareGridDim
        if rowIndex == lowestVisibleRowByColumn.get(columnIndex):
            ax.set_xlabel(f"Time ({timeUnits})")
        ax.set_ylabel("Normalized Signal")

        signalName = ":".join(signalObj.ComponentNames)
        ax.set_title(f"Group {groupIdx + 1}, Experiment {exptIdx + 1}, Signal {signalIdx + 1} [{signalName}]", fontsize=9)

    startRGB = np.array(mcolors.hex2color(startLineColor))
    endRGB = np.array(mcolors.hex2color(endLineColor))
    colorDenom = max(len(iterationList) - 1, 1)

    for iterIdx, iteration in enumerate(iterationList):
        curPlots = []
        colorFrac = iterIdx / colorDenom
        curPlotColor = mcolors.rgb2hex(startRGB + (endRGB - startRGB) * colorFrac)

        for panelIdx, (groupIdx, exptIdx, _) in enumerate(panelInfo):
            ax = flatAxes[panelIdx]
            x = evalPts[(groupIdx, exptIdx)]
            y = panelYValues[panelIdx][iterIdx]
            curPlots.append(ax.plot(x, y, color=curPlotColor, linewidth=1.8, zorder=3))

        iterTracker = addIterationTracker(fig, iteration + 1, optRes.LastIteration)

        fig.tight_layout(rect=[0, 0, 1, 0.98])
        framePath = workspaceName / f"{fileName}_{iteration + 1}.png"
        plt.savefig(framePath, dpi=100)
        framePaths.append(framePath)

        iterTracker.remove()
        if not stackImages:
            for curPlot in curPlots:
                curPlot[0].remove()

    with iio.get_writer(
        workspaceName / f"{fileName}.mp4",
        fps=fps,
        codec='libx264',
        format='FFMPEG',
        ffmpeg_params=['-pix_fmt', 'yuv420p']
    ) as writer:
        for framePath in framePaths:
            image = iio.imread(framePath)
            writer.append_data(image)

    if not keepIndividualImages:
        for framePath in framePaths:
            try:
                os.remove(framePath)
            except OSError:
                pass



############ Configuration Variables ##########
#                                             #
#               By: Colin Yancey              #
#            Created September 2025           #
#            Last Edited: 03/02/2026          #
#                  Schulman Lab               #
#            Johns Hopkins University         #
#                                             #
###############################################

# Some relevant unit conversions for CRNs.
UNIT_WEIGHTS = {
    "M": 1,
    "mM": 1e-3,
    "uM": 1e-6,
    "nM": 1e-9,

    "hours": 3600, # Avoiding h, m, s to avoid potential unit ambiguities
    "hr": 3600,
    "minutes": 60,
    "min": 60,
    "seconds": 1,
    "sec": 1
}

# Possible behavior states.
ALLOWED_BEHAVIORS = ["LOW", "HIGH"]

# Possible interruption states.
ALLOWED_INTERRUPTIONS = ["SUBSTITUTE", "ADD", "SUBTRACT"]


################ Initialization ###############
#                                             #
#               By: Colin Yancey              #
#            Created September 2025           #
#            Last Edited: 02/24/2026          #
#                  Schulman Lab               #
#            Johns Hopkins University         #
#                                             #
###############################################

from .behavior import (
    Behavior,
    BehaviorGroup,
    BehaviorTimeCourse
)

from .manager import (
   OptimizeBehaviorsManager
)


__all__ = [
    "Behavior",
    "BehaviorGroup",
    "BehaviorTimeCourse",

    "OptimizeBehaviorsManager"
]


############ EmergeX Initialization ###########
#                                             #
#               By: Colin Yancey              #
#            Created September 2025           #
#            Last Edited: 03/04/2026          #
#                  Schulman Lab               #
#            Johns Hopkins University         #
#                                             #
###############################################

# CRN Constructor and Wrappers
from emergex.crn import (
    Reaction,
    Component,
    ReactionNetwork
)

# Core
from emergex.core import (
    InitialConditionInfo,
    CRNInfoHandler,

    EvaluationPointsHandler,
    CRNSimulationRunner,

    Interruption,
    TimeSpan
)

# Optimization
from emergex.optimization import (
    CRNOptimizationFramework,
    OptimizerRunResult,

    FreeParameter,
    LinkedParameter
)

# Behaviors
from emergex.behaviors import (
    Behavior,
    BehaviorGroup,
    BehaviorTimeCourse,
    OptimizeBehaviorsManager
)

# Experiments
from emergex.experiments import (
    Signal,
    Experiment,
    ExperimentGroup,
    OptimizeExperimentsManager
)

# Results
from emergex.results import (
    CompiledOptimizationData
)

# Visualization
from emergex.visualization import (
    saveFreeParametersVisualization,
    saveBehaviorResultsVisualization,
    saveBehaviorLandscapeVisualization,
    saveExperimentResultsVisualization
)

# Utils, use these at your own risk
# Public-facing 
from emergex.utils import (
    DynamicFunctionStoring,
    getComponentHosts
)

__all__ = [
    # CRN Constructor
    "Reaction",
    "Component",
    "ReactionNetwork",

    # Core
    "InitialConditionInfo",
    "CRNInfoHandler",
    "EvaluationPointsHandler",
    "CRNSimulationRunner",
    "Interruption",
    "TimeSpan",

    # Optimization
    "CRNOptimizationFramework",
    "OptimizerRunResult",
    "FreeParameter",
    "LinkedParameter",

    # Behaviors
    "Behavior",
    "BehaviorGroup",
    "BehaviorTimeCourse",
    "OptimizeBehaviorsManager",

    # Experiments
    "Signal",
    "Experiment",
    "ExperimentGroup",
    "OptimizeExperimentsManager",

    # Results
    "CompiledOptimizationData",

    # Visualization
    "saveFreeParametersVisualization",
    "saveBehaviorResultsVisualization",
    "saveBehaviorLandscapeVisualization",
    "saveExperimentResultsVisualization",

    # Utils
    "DynamicFunctionStoring",
    "getComponentHosts"
]



################ Initialization ###############
#                                             #
#               By: Colin Yancey              #
#            Created September 2025           #
#            Last Edited: 02/24/2026          #
#                  Schulman Lab               #
#            Johns Hopkins University         #
#                                             #
###############################################

from .optimize import (
    CRNOptimizationFramework,
    OptimizerRunResult
)

from .parameters import (
    FreeParameter,
    LinkedParameter
)


__all__ = [
    "CRNOptimizationFramework",
    "OptimizerRunResult",

    "FreeParameter",
    "LinkedParameter"
]


############### CRN Optimization ##############
#                                             #
#               By: Colin Yancey              #
#            Created September 2025           #
#            Last Edited: 02/24/2026          #
#                  Schulman Lab               #
#            Johns Hopkins University         #
#                                             #
###############################################

import jax
jax.config.update("jax_enable_x64", True) # Maintain 64-bit precision.
import jax.numpy as jnp
import numpy as np

from equinox import filter_jit
import optax

from typing import Optional, Union
import time
from datetime import datetime

from .parameters import (
    FreeParameter,
    LinkedParameter
)
from ..core.crnData import (
    CRNInfoHandler,
    InitialConditionInfo
)
from ..utils.base import _getMinSecStr

###############################################

class CRNOptimizationFramework:
    def __init__(
        self,
        parameters: list[FreeParameter],
        linkedParameters: Optional[list[LinkedParameter]] = None,
        rateBoundFactor: float = 1e3,
        concentrationBoundFactor: float = 1e3,
        concentrationUnits: str = "nM",
        timeUnits: str = "sec"
    ):
        checkedRxns = []
        checkedComps = []
        if len(parameters) < 1:
            raise ValueError("No parameters were provided to the optimization framework!")
        for param in parameters:
            param.disableModifying()
            if param.ValueName == "Concentration":
                if param.Name in checkedComps:
                    raise AttributeError(f"Each free parameter must be a unique component concentration or reaction rate. The component '{param.Name}' was found to be listed at least twice.")
                checkedComps.append(param.Name)
            else:
                if (param.Name, param.ValueName) in checkedRxns:
                    raise AttributeError(f"Each free parameter must be a unique component concentration or reaction rate. The {"forward rate " if param.ValueName == "ForwardRate" else "backward rate"} of the reaction '{param.Name}' was found to be listed at least twice.")
                checkedRxns.append((param.Name, param.ValueName))
        
        self.Parameters = parameters
        self.LinkedParameters = [] if linkedParameters is None else linkedParameters

        # RateBoundFactor and ConcentrationBoundFactor define optimization search space as [initialValue / factor, initialValue * factor]
        self.RateBoundFactor = rateBoundFactor
        self.ConcentrationBoundFactor = concentrationBoundFactor
        
        self.ConcentrationUnits = concentrationUnits
        self.TimeUnits = timeUnits

        relevantBoundFactorList = [concentrationBoundFactor if param.ValueName == "Concentration" else rateBoundFactor for param in self.Parameters]

        self.LnDefaultValueList = jnp.array([jnp.log(param.DefaultValue) for param in self.Parameters])
        self.LnLowerBoundList = jnp.array([jnp.log(param.DefaultValue/relevantBoundFactorList[i]) if param.LowerBound is None else jnp.log(param.LowerBound) for i, param in enumerate(self.Parameters)])
        self.LnUpperBoundList = jnp.array([jnp.log(param.DefaultValue*relevantBoundFactorList[i]) if param.UpperBound is None else jnp.log(param.UpperBound) for i, param in enumerate(self.Parameters)])

    def adjustParameterValue(self, param: Union[FreeParameter, LinkedParameter], paramValue, crnInfo: CRNInfoHandler, rateList: jnp.array, concList: jnp.array):
        if param.ValueName in ["ForwardRate", "BackwardRate"]:
            rateList = rateList.at[crnInfo.RxnNameList.index((param.Name, param.ValueName))].set(paramValue)
        elif param.ValueName == "Concentration":
            concList = concList.at[crnInfo.CompNameList.index(param.Name)].set(paramValue)
        
        return rateList, concList

    def getStartingConditions(self, paramValues: jnp.ndarray, crnInfo: CRNInfoHandler):
        modRateList, modConcList = crnInfo.DefaultInitialConditionInfo.RateList, crnInfo.DefaultInitialConditionInfo.ConcentrationList

        for i, optParam in enumerate(self.Parameters):
            modRateList, modConcList = self.adjustParameterValue(
                optParam,
                jnp.exp(paramValues[i]),
                crnInfo,
                modRateList,
                modConcList
            )
        modRxnDict = {crnInfo.RxnNameList[i]: modRateList[i] for i in crnInfo.InterpolatedReactions}
        modCompDict = {crnInfo.CompNameList[i]: modConcList[i] for i in crnInfo.InterpolatedComponents}
        for linkedParam in self.LinkedParameters:
            modRateList, modConcList = self.adjustParameterValue(
                linkedParam,
                linkedParam.applyConversion(
                    modRxnDict,
                    modCompDict
                ),
                crnInfo,
                modRateList,
                modConcList
            )
        return InitialConditionInfo(
            modRateList,
            modConcList,
            {crnInfo.RxnNameList[i]: modRateList[i] for i in crnInfo.InterpolatedReactions},
            {crnInfo.CompNameList[i]: modConcList[i] for i in crnInfo.InterpolatedComponents}
        )
    
    def printCurrentOptimizationState(self, iteration: int, totalExpectedIterations: int, loss, iterTime, barMaxSize = 20):
        barLen = (iteration * barMaxSize) // totalExpectedIterations
        print(
            f"[{"#"*barLen + "-"*(barMaxSize - barLen)}] Iteration {iteration:4d} / {totalExpectedIterations}: Loss = {loss:.8f}\n" \
            f"  Iteration time: {iterTime:.3f} sec"
        )
    
    def optimize(self, costFn, iterationCount: int = 1000, gradThreshold: float = 1e-6, optaxOptimizerObj = None, callbackFrequency: Optional[int] = None):
        curTime = time.time()
        
        optimizer = optax.chain(
            optax.clip_by_global_norm(1.0),
            optax.adam(
                learning_rate=0.05,
                b1=0.9, # Momentum term
                b2=0.999
            ) if optaxOptimizerObj is None else optaxOptimizerObj
        )
        
        curVals = self.LnDefaultValueList
        optState = optimizer.init(curVals)
        lowerBounds, upperBounds = self.LnLowerBoundList, self.LnUpperBoundList

        @filter_jit
        def step(vals, optState):
            boundedVals = jnp.clip(vals, lowerBounds, upperBounds)
            loss, grads = jax.value_and_grad(costFn)(boundedVals)
            updates, optState = optimizer.update(grads, optState, boundedVals)
            vals = optax.apply_updates(boundedVals, updates)
            return vals, optState, loss, grads

        # Optimization loop with gradient-based and boundary-based early stopping
        losses = []
        paramHistory = []

        # Run optimization
        print("Starting optimization...")
        iterationList = np.arange(iterationCount) + 1
        callbackPts = [] if callbackFrequency is None or callbackFrequency <= 0 else [i for i in iterationList if i%callbackFrequency == 0]
        for i in iterationList:
            paramHistory.append(jnp.exp(curVals))

            iterStart = time.time()
            curVals, optState, loss, grads = step(curVals, optState)
            iterTime = time.time() - iterStart
            
            losses.append(float(loss))

            gradNorm = jnp.linalg.norm(grads)
            if gradNorm < gradThreshold:
                print(f"Converged at Iteration {i}: Gradient norm {gradNorm:.2e} is below threshold {gradThreshold:.2e}")
                break
            
            if i in callbackPts:
                self.printCurrentOptimizationState(i, iterationCount, loss, iterTime)
        
        return OptimizerRunResult(
            self,
            i,
            iterationCount,
            lossHist = losses,
            paramHist = paramHistory,
            startTime = curTime,
            finishTime = time.time()
        )


class OptimizerRunResult:
    def __init__(
        self,
        crnOptimizationFrameworkObj: CRNOptimizationFramework,
        lastIteration: int,
        iterationMax: int,
        lossHist: jnp.ndarray,
        paramHist: jnp.ndarray,
        startTime: float,
        finishTime: float
    ):
        self.CRNOptimizationFrameworkObject = crnOptimizationFrameworkObj

        self.LastIteration = lastIteration
        self.IterationMaximum = iterationMax

        self.LossHistory = lossHist
        self.ParameterHistory = paramHist

        self.StartTime = startTime
        self.FinishTime = finishTime

        self.StartDateTime = datetime.fromtimestamp(self.StartTime)
        self.FinishDateTime = datetime.fromtimestamp(self.FinishTime)

        self.TimeDelta = self.FinishTime - self.StartTime

        self.OutputStr = f"\n{"="*60}\n{" "*19}OPTIMIZATION RESULTS\n{"="*60}\n" \
        f"Optimization Time: {_getMinSecStr(self.TimeDelta)}\n" \
        f"Final Optimization Iteration: {self.LastIteration} / {self.IterationMaximum}\n\n" \
        f"Initial Loss: {self.LossHistory[0]:.4g}\n" \
        f"Final Loss: {self.LossHistory[-1]:.4g}\n" \
        f"Loss Reduction: {((self.LossHistory[0] - self.LossHistory[-1]) / self.LossHistory[0] * 100):.4g}%\n" \
        f"\nPARAMETER CHANGES:\n" \
        f"{"-"*40}\n"
        
        for i, param in enumerate(self.CRNOptimizationFrameworkObject.Parameters):
            iP = self.ParameterHistory[0][i]
            fP = self.ParameterHistory[-1][i]
            change = ((fP - iP) / iP * 100)
            
            self.OutputStr += f"{param}:\n" \
            f"  Initial: {iP:.4g}\n" \
            f"  Final:   {fP:.4g}\n" \
            f"  Change:  {change:+.2f}%\n"
    
    def __str__(self):
        return self.OutputStr



################ Initialization ###############
#                                             #
#               By: Colin Yancey              #
#             Created October 2025            #
#            Last Edited: 02/24/2026          #
#                  Schulman Lab               #
#            Johns Hopkins University         #
#                                             #
###############################################


from .optimizerResults import (
    CompiledOptimizationData
)


__all__ = [
    "CompiledOptimizationData"
]


################ Initialization ###############
#                                             #
#               By: Colin Yancey              #
#             Created January 2026            #
#            Last Edited: 02/24/2026          #
#                  Schulman Lab               #
#            Johns Hopkins University         #
#                                             #
###############################################


from .elements import (
    addIterationTracker
)

# from .iterations import (
#     getIterationList
# )

__all__ = [
    # Figure elements
    "addIterationTracker",
    
    # Iteration parsing handler
    # "getIterationList"
]


########## Optimizer Results Handling #########
#                                             #
#               By: Colin Yancey              #
#             Created October 2025            #
#            Last Edited: 02/24/2026          #
#                  Schulman Lab               #
#            Johns Hopkins University         #
#                                             #
###############################################

from ..utils.base import DataStore
from ..utils.helpers import ManagerClass
from ..core.simulate import CRNSimulationRunner
from ..optimization.optimize import CRNOptimizationFramework

###############################################

# Class for handling all relevant context needed to uniquely identify an optimization run result (the input parameters that
# led to the result generation and the result itself). Can save optimization runs, and can load previously saved runs.
class CompiledOptimizationData(DataStore):
    def __init__(
        self,
        managerObj: ManagerClass,
        simulationRunnerObj: CRNSimulationRunner,
        optimizerFramework: CRNOptimizationFramework,
        iterationCount: int = 1000,
        callbackFrequency: int = None
    ):
        self.ManagerObject = managerObj
        self.SimulationRunnerObject = simulationRunnerObj

        managerCostFn = managerObj.getCostFn(optimizerFramework, simulationRunnerObj)
        optimizedResult = optimizerFramework.optimize(managerCostFn, iterationCount = iterationCount, callbackFrequency = callbackFrequency)

        self.OptimizerRunResultObject = optimizedResult



################ Initialization ###############
#                                             #
#               By: Colin Yancey              #
#            Created September 2025           #
#            Last Edited: 02/24/2026          #
#                  Schulman Lab               #
#            Johns Hopkins University         #
#                                             #
###############################################

from .experiment import (
    Signal,
    Experiment,
    ExperimentGroup
)

# from .experimentCompiler import (

# )

from .manager import (
   OptimizeExperimentsManager
)


__all__ = [
    "Signal",
    "Experiment",
    "ExperimentGroup",

    "OptimizeExperimentsManager"
]


################ CRN Parameters ###############
#                                             #
#               By: Colin Yancey              #
#            Created September 2025           #
#            Last Edited: 02/24/2026          #
#                  Schulman Lab               #
#            Johns Hopkins University         #
#                                             #
###############################################

from typing import Literal, Optional, Callable

from ..crn import (
    Reaction,
    Component,
    _getRxnStr
)
from ..utils.base import (
    _getValueName,
    _getParameterName,
    DynamicFunctionStoring
)

###############################################


class FreeParameter:
    def __init__(
        self,
        parameterObject: Component | Reaction,
        valueName: Optional[Literal["ForwardRate", "BackwardRate", "Concentration"]] = None,
        lowerBound: float = None,
        upperBound: float = None
    ):
        self.ValueName = _getValueName(parameterObject, valueName)
        self.ParameterObject = parameterObject

        if isinstance(parameterObject, Component):
            self.Name = parameterObject.Name
            self.DefaultValue = parameterObject.Concentration
            self.StrRep = self.Name
        else:
            self.Name = parameterObject.RxnStr
            self.DefaultValue = parameterObject.ForwardRate if self.ValueName == "ForwardRate" else parameterObject.BackwardRate
            if self.ValueName == "ForwardRate": # Orient the direction of the string output to match the rate of the free parameter in question.
                subs, prods = parameterObject.Substrates, parameterObject.Products
            else:
                subs, prods = parameterObject.Products, parameterObject.Substrates
            self.StrRep = _getRxnStr(subs, prods, backwardRate = False)

        self.LowerBound = lowerBound
        self.UpperBound = upperBound
        self._Modifiable = True
    
    def tryModifying(self):
        if not self._Modifiable:
            raise ValueError("This free parameter is no longer modifiable. Modifications must be made before creating dependence on this object.")

    def changeDefaultValue(self, newVal):
        self.tryModifying()
        if (self.LowerBound and self.LowerBound > newVal) or (self.UpperBound and self.UpperBound < newVal):
            raise ValueError(f"Attempted to set default value of {self.Name} to {self.DefaultValue:.4g}, outside of provided bound(s).")
        else:
            self.DefaultValue = newVal
    
    def disableModifying(self):
        self._Modifiable = False

    def __str__(self):
        return self.StrRep


# Assign a relationship between a parameter and the rest of the reaction network's initial conditions (by default, it simply clones its reference's relevant value of interest).
class LinkedParameter(DynamicFunctionStoring):
    def __init__(
        self,
        parameterObject: Component | Reaction,
        valueName: Optional[Literal["ForwardRate", "BackwardRate", "Concentration"]] = None,
        refName: Optional[str] = None,
        conversionFn: Optional[Callable[[dict, dict], float | int]] = None
    ):
        self.ValueName = _getValueName(parameterObject, valueName)
        self.Name = _getParameterName(parameterObject)

        if refName != None and conversionFn is None:
            self.ReferenceName = refName
            if self.ValueName in ["ForwardRate", "BackwardRate"]:
                self.applyConversion = self._convertRates
            else:
                self.applyConversion = self._convertConcs
        elif conversionFn != None:
            self.applyConversion = conversionFn
        else:
            raise ValueError(f"Must provide either a reference name for the linked parameter '{self.Name}' or a custom conversion function, cannot be both or neither.")
        
    def _convertRates(self, rates, _):
        return rates[(self.ReferenceName, self.ValueName)]
    
    def _convertConcs(self, _, concs):
        return concs[self.ReferenceName]



################ Initialization ###############
#                                             #
#               By: Colin Yancey              #
#            Created September 2025           #
#            Last Edited: 02/24/2026          #
#                  Schulman Lab               #
#            Johns Hopkins University         #
#                                             #
###############################################

from .crnData import (
    InitialConditionInfo,
    CRNInfoHandler 
)

from .simulate import (
    EvaluationPointsHandler,
    CRNSimulationRunner
)

from .timeCourse import (
    Interruption,
    TimeSpan
)


__all__ = [
    "InitialConditionInfo",
    "CRNInfoHandler",

    "EvaluationPointsHandler",
    "CRNSimulationRunner",

    "Interruption",
    "TimeSpan"
]


################ Figure Elements ##############
#                                             #
#               By: Colin Yancey              #
#             Created January 2026            #
#            Last Edited: 02/24/2026          #
#                  Schulman Lab               #
#            Johns Hopkins University         #
#                                             #
###############################################

from matplotlib.figure import Figure as figureClass

###############################################

def addIterationTracker(fig: figureClass, iteration: int, totalIter: int, posX: float = 0.5, posY: float = 0.99):
    txt = fig.text(
        posX,
        posY,
        f"Iteration {iteration} / {totalIter}",
        transform=fig.transFigure,
        fontsize=10,
        ha='center',
        va='top',
        bbox=dict(facecolor='white', alpha=0.5)
    )
    return txt


####### Behavior Optimization Visualizer ######
#                                             #
#               By: Colin Yancey              #
#             Created January 2026            #
#            Last Edited: 02/24/2026          #
#                  Schulman Lab               #
#            Johns Hopkins University         #
#                                             #
###############################################

from pathlib import Path
from typing import Optional
import numpy as np
import jax.numpy as jnp
import time

import imageio.v2 as iio
import os

import matplotlib.pyplot as plt
from matplotlib import colors as mcolors

from ..results.optimizerResults import CompiledOptimizationData
from ..behaviors.manager import OptimizeBehaviorsManager
from ..utils.constants import UNIT_WEIGHTS
from ..utils.base import _getMinSecStr
from ..utils.helpers import _checkManagerType
from .plotting import *

from ..core.simulate import EvaluationPointsHandler

###############################################

# Returns an animated visualization of the tracked behavior functions and how they changed over the optimization process.
def saveBehaviorResultsVisualization(
    compiledOptDataObject: CompiledOptimizationData,
    timeCourseBehaviorGroupWhitelist: Optional[list[int | tuple[int, int]]] = None,
    iterationList: Optional[list[int]] = None, # If provided, plot will only contain the selected iterations in the list, each denoted by either a number indicating the time course number, or a tuple containing the time course number and the behavior group number within that time course.
    fileLocation: Path = Path('./'),
    fileName: str = 'behaviorResults',
    fps: int = 1,
    stackImages: bool = False,
    keepIndividualImages: bool = False,
    startLineColor = "#202020",
    endLineColor = "#30E0F0",
    timeUnits = "sec",
    dataResolution: int = 1001
):
    _checkManagerType(compiledOptDataObject.ManagerObject, OptimizeBehaviorsManager)
    
    timeCourses = compiledOptDataObject.ManagerObject.BehaviorTimeCourses
    simRunner = compiledOptDataObject.SimulationRunnerObject
    crnInfo = compiledOptDataObject.ManagerObject.CRNInfo
    optRes = compiledOptDataObject.OptimizerRunResultObject
    optFrameworkObj = optRes.CRNOptimizationFrameworkObject

    totalIter = optRes.LastIteration
    if iterationList is None:
        iterationList = np.arange(totalIter)
    else:
        iterationList = np.array(iterationList) - 1

    if len(iterationList) == 0:
        raise ValueError("iterationList must contain at least one iteration.")
    if np.any(iterationList < 0) or np.any(iterationList >= totalIter):
        raise ValueError(f"All entries in iterationList must be between 1 and {totalIter}, inclusive.")
    
    timeCoursesLen = len(timeCourses)

    def raiseTimeCourseLenError(selectionNum):
        raise ValueError(f"Since only {timeCoursesLen} time courses were provided in this optimization run, Time Course {selectionNum} does not exist.")

    timeCoursesToSolve = set()
    timeCourseSolutions = [0]*timeCoursesLen
    if timeCourseBehaviorGroupWhitelist is None:
        timeCourseBehaviorGroupWhitelist = []
        for i, timeCourse in enumerate(timeCourses):
            timeCoursesToSolve.add(i)
            for j in range(len(timeCourse.BehaviorGroups)):
                timeCourseBehaviorGroupWhitelist.append((i + 1, j + 1))
    elif len(timeCourseBehaviorGroupWhitelist) == 0:
        raise ValueError("If providing a whitelist, the behavior evolution visualization must contain at least one tuple indicating which time course's behavior group to plot.")
    else:
        for i, selection in timeCourseBehaviorGroupWhitelist:
            if isinstance(selection, int):
                newSelection = selection - 1
                timeCoursesToSolve.add(newSelection)
            else:
                newSelection = selection[0] - 1
                timeCoursesToSolve.add(newSelection)
                if selection[0] < 1 or selection[0] > timeCoursesLen:
                    raiseTimeCourseLenError(selection[0])
                if selection[1] < 1 or selection[1] > len(timeCourses[newSelection].BehaviorGroups):
                    raise ValueError(f"Behavior Group {selection[1]} is not contained within Time Course {selection[0]}")
    
    evalPts = [0]*timeCoursesLen
    evalPtsParsed = [0]*timeCoursesLen

    for timeCourseNum in timeCoursesToSolve:
        if timeCourseNum < 0 or timeCourseNum >= timeCoursesLen:
            raiseTimeCourseLenError(timeCourseNum + 1)
        
        evalPts[timeCourseNum] = np.linspace(0, sum([timeSpan.Time for timeSpan in timeCourses[timeCourseNum].TimeCourse]), dataResolution)
        evalPtsParsed[timeCourseNum] = EvaluationPointsHandler(evalPts[timeCourseNum]).parseEvalPts(timeCourses[timeCourseNum].TimeCourse)

    print(f"Computing integration solutions...\n{"#"*40}")
    startTime = time.time()

    initCondComps = [None] * len(iterationList)

    firstIter = iterationList[0]
    initCondInfo = optFrameworkObj.getStartingConditions(jnp.log(optRes.ParameterHistory[firstIter]), crnInfo)
    initCondComps[0] = initCondInfo.ComponentsDictionary
    for timeCourseNum in timeCoursesToSolve:
        timeCourseSolutions[timeCourseNum] = [
            simRunner.runTimeCourse(evalPtsParsed[timeCourseNum], crnInfo, initCondInfo)
        ]
    
    iterTime = time.time() - startTime
    completionTimeEstimate = iterTime*(len(iterationList)-1)
    print(
        "Integration solutions for the first iteration of the chosen time course behavior group pairs have been found.\n" \
        f"  Time: {iterTime:.3f} sec\n" \
        f"  Estimated time to completion: {_getMinSecStr(completionTimeEstimate)}"
    )
    if completionTimeEstimate > 120:
        print(f"  If this is too long, it is recommended to use a smaller sample of time courses or a sparser iteration sampling.")
    
    for iterIdx, iter in enumerate(iterationList[1:], start = 1):
        initCondInfo = optFrameworkObj.getStartingConditions(jnp.log(optRes.ParameterHistory[iter]), crnInfo)
        initCondComps[iterIdx] = initCondInfo.ComponentsDictionary
        for timeCourseNum in timeCoursesToSolve:
            timeCourseSolutions[timeCourseNum].append(
                simRunner.runTimeCourse(evalPtsParsed[timeCourseNum], crnInfo, initCondInfo)
            )
    
    print(
        "All integration solutions found.\n" \
        f"  Total Time (including first iteration): {_getMinSecStr(time.time() - startTime)}\n"
        "Creating plots..."
    )

    # Prepare output workspace
    workspaceName = fileLocation / 'mp4 Output'
    os.makedirs(workspaceName, exist_ok=True)
    framePaths = []

    startRGB = np.array(mcolors.hex2color(startLineColor))
    endRGB = np.array(mcolors.hex2color(endLineColor))
    startIteration, endIteration = iterationList[0], iterationList[-1]
    
    panelNum = len(timeCourseBehaviorGroupWhitelist)
    squareGridDim = int(np.ceil(np.sqrt(panelNum))) # Square grid of the panel count
    gridLen = 4*squareGridDim
    lowMax = compiledOptDataObject.ManagerObject.getLowMax()
    highMin = compiledOptDataObject.ManagerObject.getHighMin()

    fig, axes = plt.subplots(squareGridDim, squareGridDim, figsize=(gridLen, gridLen))
    if squareGridDim == 1: # Fix annoying auto-axis size leading to dimension reduction
        axes = np.array([[axes]])
    flatAxes = axes.ravel()

    for m in range(panelNum, squareGridDim**2):
        flatAxes[m].axis('off')

    lowestVisibleRowByColumn = {}
    for columnIndex in range(squareGridDim):
        visibleRows = [
            rowIndex
            for rowIndex in range(squareGridDim)
            if (rowIndex * squareGridDim + columnIndex) < panelNum
        ]
        if len(visibleRows) > 0:
            lowestVisibleRowByColumn[columnIndex] = max(visibleRows)

    for k, panelInfo in enumerate(timeCourseBehaviorGroupWhitelist):
        tCIdx = panelInfo[0] - 1
        timeCourseObj = timeCourses[tCIdx]
        x = evalPts[tCIdx]/UNIT_WEIGHTS[timeUnits]

        ax = flatAxes[k]
        missingLowLabel = True
        missingHighLabel = True

        for behavior in timeCourseObj.BehaviorGroups[panelInfo[1] - 1].Behaviors:
            startBound = behavior.StartBound/UNIT_WEIGHTS[timeUnits]
            endBound = (behavior.EndBound if behavior.EndBound != float('inf') else sum([ts.Time for ts in timeCourseObj.TimeCourse]))/UNIT_WEIGHTS[timeUnits]
            
            curLabel = None
            if behavior.BehaviorType == "LOW" and missingLowLabel:
                curLabel = "LOW target"
                missingLowLabel = False
            elif behavior.BehaviorType == "HIGH" and missingHighLabel:
                curLabel = "HIGH target"
                missingHighLabel = False
            
            if behavior.BehaviorType == "LOW":
                ax.plot([startBound, endBound], [lowMax, lowMax], color = "#FF4040", linestyle = '--', alpha = 0.7, label = 'LOW threshold')
                ax.axvspan(startBound, endBound, ymax = (lowMax + 0.03) / 1.06, alpha = 0.2, color = "#FF4040", label = curLabel)
            elif behavior.BehaviorType == "HIGH":
                ax.plot([startBound, endBound], [highMin, highMin], color = "#228B22", linestyle='--', alpha=0.7, label = 'HIGH threshold')
                ax.axvspan(startBound, endBound, ymin = (highMin + 0.03) / 1.06, alpha = 0.2, color = "#228B22", label = curLabel)

        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)
        ax.grid(True, linestyle='--', alpha = 0.3)
        ax.set_xlim(x[0], x[-1])
        rowIndex = k // squareGridDim
        columnIndex = k % squareGridDim
        if rowIndex == lowestVisibleRowByColumn.get(columnIndex):
            ax.set_xlabel(f"Time ({timeUnits})")

        ax.set_ylim(-0.03, 1.03)

    panelYValues = [[None] * len(iterationList) for _ in range(panelNum)]
    for k, panelInfo in enumerate(timeCourseBehaviorGroupWhitelist):
        tCIdx = panelInfo[0] - 1
        behaviorGroup = timeCourses[tCIdx].BehaviorGroups[panelInfo[1] - 1]
        relevantIndices = behaviorGroup.getRelevantComponentIndices(crnInfo.CompNameList)

        for iterIdx in range(len(iterationList)):
            relevantResults = timeCourseSolutions[tCIdx][iterIdx][:, relevantIndices]
            if len(relevantIndices) == 1:
                rawSignalVals = relevantResults.reshape(-1)
            else:
                rawSignalVals = jnp.sum(relevantResults, axis = 1)

            panelYValues[k][iterIdx] = behaviorGroup.getNormalizeVMAP(initCondComps[iterIdx])(rawSignalVals)

    # Generate a frame for each iteration in iterationList
    for iterIdx, iteration in enumerate(iterationList):
        curPlots = []
        for k, panelInfo in enumerate(timeCourseBehaviorGroupWhitelist):
            tCIdx = panelInfo[0] - 1
            timeCourseObj = timeCourses[tCIdx]
            ax = flatAxes[k]
            x = evalPts[tCIdx]/UNIT_WEIGHTS[timeUnits]
            behaviorGroup = timeCourseObj.BehaviorGroups[panelInfo[1] - 1]
            y = panelYValues[k][iterIdx]
            curPlotColor = mcolors.rgb2hex(startRGB + (endRGB - startRGB) * ((iteration - startIteration) / (endIteration - startIteration + 1)))
            curPlots.append(
                ax.plot(x, y, color = curPlotColor, linewidth = 1.8)
            )
            ax.set_title(fr"Time Course {tCIdx + 1}, $\mathbf{{f}}\left(\text{{[{behaviorGroup.ComponentNames[0]}]}}\right)$", fontsize=9)

        iterTracker = addIterationTracker(fig, iteration + 1, optRes.LastIteration)

        fig.tight_layout(rect = [0, 0, 1, 0.98])
        framePath = workspaceName / f"{fileName}_{iteration}.png"
        plt.savefig(framePath, dpi=100)
        framePaths.append(framePath)
        
        iterTracker.remove()
        if not stackImages:
            for curPlot in curPlots:
                curPlot[0].remove()

    # Write MP4
    with iio.get_writer(workspaceName / f"{fileName}.mp4", fps = fps, codec = 'libx264', format = 'FFMPEG', ffmpeg_params = ['-pix_fmt', 'yuv420p']) as writer:
        for framePath in framePaths:
            image = iio.imread(framePath)
            writer.append_data(image)

    if not keepIndividualImages:
        for framePath in framePaths:
            try:
                os.remove(framePath)
            except OSError:
                pass


def saveBehaviorLandscapeVisualization(
    optBehaviorManagerObject: OptimizeBehaviorsManager,
    timeCourseBehaviorGroupWhitelist: Optional[list[int | tuple[int, int]]] = None,
    fileLocation: Path = Path('./'),
    fileName: str = 'behaviorLandscapeScaffold',
    timeUnits: str = "sec"
):
    if timeUnits not in UNIT_WEIGHTS:
        raise ValueError(f"Unknown timeUnits '{timeUnits}'. Available units: {list(UNIT_WEIGHTS.keys())}")

    timeCourses = optBehaviorManagerObject.BehaviorTimeCourses
    timeCoursesLen = len(timeCourses)

    def raiseTimeCourseLenError(selectionNum):
        raise ValueError(f"Since only {timeCoursesLen} time courses were provided in this optimization run, Time Course {selectionNum} does not exist.")

    panelInfo = []
    if timeCourseBehaviorGroupWhitelist is None:
        for i, timeCourseObj in enumerate(timeCourses):
            for j in range(len(timeCourseObj.BehaviorGroups)):
                panelInfo.append((i + 1, j + 1))
    elif len(timeCourseBehaviorGroupWhitelist) == 0:
        raise ValueError("If providing a whitelist, it must contain at least one entry.")
    else:
        for selection in timeCourseBehaviorGroupWhitelist:
            if isinstance(selection, int):
                timeCourseIdx = selection - 1
                if timeCourseIdx < 0 or timeCourseIdx >= timeCoursesLen:
                    raiseTimeCourseLenError(selection)
                for behaviorGroupIdx in range(len(timeCourses[timeCourseIdx].BehaviorGroups)):
                    panelInfo.append((selection, behaviorGroupIdx + 1))
            else:
                if len(selection) != 2:
                    raise ValueError("Each whitelist tuple must contain exactly (timeCourseNumber, behaviorGroupNumber).")
                timeCourseNum, behaviorGroupNum = selection
                timeCourseIdx = timeCourseNum - 1
                if timeCourseIdx < 0 or timeCourseIdx >= timeCoursesLen:
                    raiseTimeCourseLenError(timeCourseNum)
                if behaviorGroupNum < 1 or behaviorGroupNum > len(timeCourses[timeCourseIdx].BehaviorGroups):
                    raise ValueError(f"Behavior Group {behaviorGroupNum} is not contained within Time Course {timeCourseNum}.")
                panelInfo.append((timeCourseNum, behaviorGroupNum))

    panelNum = len(panelInfo)
    if panelNum == 0:
        raise ValueError("No behavior panels selected to scaffold.")

    squareGridDim = int(np.ceil(np.sqrt(panelNum)))
    gridLen = 4 * squareGridDim
    lowMax = optBehaviorManagerObject.getLowMax()
    highMin = optBehaviorManagerObject.getHighMin()

    fig, axes = plt.subplots(squareGridDim, squareGridDim, figsize=(gridLen, gridLen))
    if squareGridDim == 1:
        axes = np.array([[axes]])
    flatAxes = axes.ravel()

    for m in range(panelNum, squareGridDim**2):
        flatAxes[m].axis('off')

    lowestVisibleRowByColumn = {}
    for columnIndex in range(squareGridDim):
        visibleRows = [
            rowIndex
            for rowIndex in range(squareGridDim)
            if (rowIndex * squareGridDim + columnIndex) < panelNum
        ]
        if len(visibleRows) > 0:
            lowestVisibleRowByColumn[columnIndex] = max(visibleRows)

    for panelIdx, panelPair in enumerate(panelInfo):
        timeCourseIdx = panelPair[0] - 1
        behaviorGroupIdx = panelPair[1] - 1
        timeCourseObj = timeCourses[timeCourseIdx]
        behaviorGroupObj = timeCourseObj.BehaviorGroups[behaviorGroupIdx]

        totalTime = sum([timeSpan.Time for timeSpan in timeCourseObj.TimeCourse]) / UNIT_WEIGHTS[timeUnits]
        ax = flatAxes[panelIdx]

        missingLowLabel = True
        missingHighLabel = True
        for behavior in behaviorGroupObj.Behaviors:
            startBound = behavior.StartBound / UNIT_WEIGHTS[timeUnits]
            endBound = (behavior.EndBound if behavior.EndBound != float('inf') else sum([ts.Time for ts in timeCourseObj.TimeCourse])) / UNIT_WEIGHTS[timeUnits]

            curLabel = None
            if behavior.BehaviorType == "LOW" and missingLowLabel:
                curLabel = "LOW target"
                missingLowLabel = False
            elif behavior.BehaviorType == "HIGH" and missingHighLabel:
                curLabel = "HIGH target"
                missingHighLabel = False

            if behavior.BehaviorType == "LOW":
                ax.plot([startBound, endBound], [lowMax, lowMax], color="#FF4040", linestyle='--', alpha=0.7, label='LOW threshold')
                ax.axvspan(startBound, endBound, ymax=(lowMax + 0.03) / 1.06, alpha=0.2, color="#FF4040", label=curLabel)
            elif behavior.BehaviorType == "HIGH":
                ax.plot([startBound, endBound], [highMin, highMin], color="#228B22", linestyle='--', alpha=0.7, label='HIGH threshold')
                ax.axvspan(startBound, endBound, ymin=(highMin + 0.03) / 1.06, alpha=0.2, color="#228B22", label=curLabel)

        ax.spines['top'].set_visible(False)
        ax.spines['right'].set_visible(False)
        ax.grid(True, linestyle='--', alpha=0.3)
        ax.set_xlim(0, totalTime)
        ax.set_ylim(-0.03, 1.03)

        rowIndex = panelIdx // squareGridDim
        columnIndex = panelIdx % squareGridDim
        if rowIndex == lowestVisibleRowByColumn.get(columnIndex):
            ax.set_xlabel(f"Time ({timeUnits})")

        ax.set_title(
            fr"Time Course {timeCourseIdx + 1}, $\mathbf{{f}}\left(\text{{[{behaviorGroupObj.ComponentNames[0]}]}}\right)$",
            fontsize=9
        )

    fig.tight_layout(rect=[0, 0, 1, 0.98])
    os.makedirs(fileLocation, exist_ok=True)
    plt.savefig(fileLocation / f"{fileName}.png", dpi=100)
    plt.close(fig)




############## Experiment Manager #############
#                                             #
#               By: Colin Yancey              #
#            Created September 2025           #
#            Last Edited: 02/24/2026          #
#                  Schulman Lab               #
#            Johns Hopkins University         #
#                                             #
###############################################

import jax.numpy as jnp

from ..core.simulate import (
    CRNSimulationRunner
)
from ..optimization.optimize import (
    CRNOptimizationFramework
)
from ..utils.helpers import ManagerClass
from .experiment import ExperimentGroup

###############################################

class OptimizeExperimentsManager(ManagerClass):
    def __init__(self, experimentGroups: list[ExperimentGroup]):
        self.ExperimentGroups = experimentGroups

        self.NormalizedWeights = jnp.array([exptGroup.Weight for exptGroup in self.ExperimentGroups])
        self.NormalizedWeights /= jnp.sum(self.NormalizedWeights)

    def getCostFn(self, optimizerFrameworkObj: CRNOptimizationFramework, simulationRunnerObj: CRNSimulationRunner):
        exptGroups = self.ExperimentGroups
        normWeights = self.NormalizedWeights
        
        precomputedTimeCourseData = []
        for experimentGroupObj in self.ExperimentGroups:
            timePointHandlerObj = experimentGroupObj.TimePointsHandler
            precomputedTimeCourseData.append([
                (
                    timePointHandlerObj.parseEvalPts(expt.TimeCourse),
                    jnp.arange(len(timePointHandlerObj.TotalEvalPts))
                ) for expt in experimentGroupObj.Experiments
            ])
        
        def getExperimentFitCost(y):

            exptCost = 0
            for i, experimentGroupObj in enumerate(exptGroups):
                crnInfo = experimentGroupObj.CRNInfo
                initCondInfo = optimizerFrameworkObj.getStartingConditions(y, crnInfo)
                
                for j, expt in enumerate(experimentGroupObj.Experiments):
                    results = simulationRunnerObj.runTimeCourse(precomputedTimeCourseData[i][j][0], crnInfo, initCondInfo)

                    signalCost = 0.0
                    for k, signal in enumerate(expt.Signals):
                        resultOfInterest = signal.getNormalizeVMAP(initCondInfo.ComponentsDictionary)(
                            jnp.sum(
                                jnp.atleast_2d(
                                    results[precomputedTimeCourseData[i][j][1], signal.getRelevantComponentIndices(crnInfo.CompNameList)]
                                ),
                                axis=0
                            )
                        )
                        signalCost += jnp.sum(jnp.square(resultOfInterest - signal.DataPoints) * expt.SignalWeights[k])
                    
                    exptCost += signalCost * experimentGroupObj.ExperimentWeights[j] * normWeights[i]
            
            return exptCost
        
        return getExperimentFitCost



######## Chemical Reaction Network Data #######
#                                             #
#               By: Colin Yancey              #
#            Created September 2025           #
#            Last Edited: 02/24/2026          #
#                  Schulman Lab               #
#            Johns Hopkins University         #
#                                             #
###############################################

from dataclasses import dataclass

import jax.numpy as jnp

from ..crn import ReactionNetwork

###############################################

@dataclass
class InitialConditionInfo:
    RateList: jnp.array
    ConcentrationList: jnp.array
    ReactionsDictionary: dict
    ComponentsDictionary: dict

class CRNInfoHandler:
    def __init__(self, crn: ReactionNetwork):
        self.CRN = crn
        
        # Need to keep track of all rates, even ones initially set to zero, incase there is a linked parameter or interruption which modifies a zero backward rate.
        compNameList, rxnNameList, rateList, rxnMatrix, modifierMatrix = self.CRN.getCompiledCRNMatricies(pruneDisabledBackward = False)
        crn.disableModifying()

        self.CompNameList = compNameList
        self.RxnNameList = rxnNameList
        self.ReactionMatrix = jnp.array(rxnMatrix)
        self.ModifierMatrixTransposed = jnp.array(modifierMatrix).T

        self.InterpolatedReactions = range(len(rxnNameList))
        self.InterpolatedComponents = range(len(compNameList))

        rateList = jnp.array(rateList)
        concList = jnp.array([self.CRN.Components[compName].Concentration for compName in compNameList])

        self.DefaultInitialConditionInfo = InitialConditionInfo(
            rateList,
            concList,
            {rxnNameList[i]: rateList[i] for i in self.InterpolatedReactions},
            {compNameList[i]: concList[i] for i in self.InterpolatedComponents}
        )
    
    def convertTimeCourseResultToDict(self, results: jnp.array):
        return {compName: results[:, compIdx] for compIdx, compName in enumerate(self.CompNameList)}


################ Initialization ###############
#                                             #
#               By: Colin Yancey              #
#             Created January 2026            #
#            Last Edited: 03/04/2026          #
#                  Schulman Lab               #
#            Johns Hopkins University         #
#                                             #
###############################################

from .observeParameters import (
    saveFreeParametersVisualization
)

from .observeBehaviors import (
    saveBehaviorResultsVisualization,
    saveBehaviorLandscapeVisualization
)

from .observeExperiments import (
    saveExperimentResultsVisualization
)

__all__ = [
    # Observe parameters
    "saveFreeParametersVisualization",

    # Observe behaviors
    "saveBehaviorResultsVisualization",
    "saveBehaviorLandscapeVisualization",

    # Observe experiments
    "saveExperimentResultsVisualization"
]



#### Chemical Reaction Network Constructor ####
#                                             #
#               By: Colin Yancey              #
#             Created February 2025           #
#            Last Edited: 03/05/2026          #
#                 Schulman Lab                #
#           Johns Hopkins University          #
#                                             #
###############################################

from typing import Optional
import re
import numpy as np
import collections
import scipy.integrate as spi
import warnings

# Strings forbidden in component names due to their use in reaction string parsing or their potential inconsistent omission.
PREVENTED_STRINGS = [" ", ",", "<-", "->", "+"]

###############################################

def _getRxnStr(substrates: list, products: list, backwardRate: bool = False):
    """
    [Internal] Generate a standardized reaction string from substrates and products.
    
    Args:
        substrates: List of substrate names.
        products: List of product names.
        backwardRate: Determines arrow used ("<->" or "->").
    
    Returns:
        A formatted reaction string (e.g., "A + B -> C" or "A + B <-> C").
    """
    substrates.sort()
    products.sort()
    return " + ".join(substrates) + (" <-> " if backwardRate else " -> ") + " + ".join(products)


def _getSubstratesAndProducts(rxnStr: str, backwardRate: bool = False):
    """
    [Internal] Parse a reaction string to extract substrates and products.
    
    Args:
        rxnStr: A reaction string (e.g., "A + B -> C" or "A + B <-> C").
        backwardRate: Determines arrow used ("<->" or "->").
    
    Returns:
        Two lists, one for substrates and one for products.
    """
    if backwardRate:
        eqnParse = re.match(r"(.+)<->(.+)", rxnStr)
    else:
        eqnParse = re.match(r"(.+)->(.*)", rxnStr)
    
    if eqnParse:
        return eqnParse.group(1).split("+"), eqnParse.group(2).split("+")
    else:
        raise ValueError("The chemical equation is not in the correct format.")


class Component:
    """
    Creates a Component object representing an item in a solution. Has a name and a concentration.

    -- Functions --
    __init__ (Initializing function)
        Initializes a Component. Must provide a name and concentration.


    -- Properties --
    Name (IMMUTABLE): Name of the Component.
    Concentration: Concentration of the Component.
    """
    def __init__(self, compName, compConc):
        if any(char in compName for char in PREVENTED_STRINGS): # Checks for forbidden characters
            raise ValueError(f"Avoid using spaces, commas, +, <-, and -> characters in component names. Violating name: '{compName}'")

        self._Name = compName
        self.Concentration = compConc
    
    @property
    def Name(self):
        return self._Name

    def __str__(self):
        return self.Name


class Reaction:
    """
    Creates a Reaction object representing a chemical reaction. Has a forward rate and a backward rate, as well as a list of substrates and products.

    -- Functions --
    __init__ (Initializing function)
        Initializes a Reaction. Must provide a forward rate and either a reaction string (i.e., "A + B -> C") or substrate and product lists (i.e., ["A", "B"], ["C"]).


    -- Properties --
    ForwardRate: The forward rate of the reaction (i.e., rate at which C is produced in A + B -> C).
    BackwardRate: The optional backward rate of the reaction (i.e., the rate at which C is consumed in A + B <-> C, independent of the ForwardRate).
    Substrates (IMMUTABLE): The substrates in the reaction.
    Products (IMMUTABLE): The products in the reaction.
    RxnStr (IMMUTABLE): The reaction string corresponding to the defined reaction. Substrates and products may be reordered from how they were provided.


    -- NOTES --
    Stoichiometry must be given without the use of coefficients, as the coefficients will be interpretted as part of the unique string. To implement
    stoichiometry (i.e., 4A + 2B -> C), add multiple of the same substrate or product together in the provided string (i.e., A + A + A + A + B + B -> C),
    or the provided substrate or product table (i.e., ["A", "A", "A", "A", "B", "B"])
    """
    def __init__(self, forwardRate: float, rxnStr: Optional[str] = None, backwardRate: float = 0.0, substrates: Optional[list] = [], products: Optional[list] = []):
        if forwardRate < 0 or backwardRate < 0:
            raise ValueError("Rates must not be negative.")
        
        if rxnStr:
            substrates, products = _getSubstratesAndProducts(rxnStr, backwardRate != 0)
        
        if len(substrates)==0:
            raise ValueError("Missing substrates; each reaction must have a substrate.")
        
        substrates = [x.strip() for x in substrates]
        if len(products) > 0:
            products = [x.strip() for x in products]
        
        if collections.Counter(substrates) == collections.Counter(products):
            raise ValueError("This reaction appears to do nothing. Check your substrates and products.")
        
        self.ForwardRate = forwardRate
        self.BackwardRate = backwardRate
        self._RxnStr = _getRxnStr(substrates, products, backwardRate > 0.0)
        self._Substrates = substrates
        self._Products = products

    @property
    def RxnStr(self):
        return self._RxnStr
    
    @property
    def Substrates(self):
        return self._Substrates
    
    @property
    def Products(self):
        return self._Products
    
    # Returns the reaction string
    def __str__(self):
        return self.RxnStr


def _compareReactions(rxn1: Reaction, rxn2: Reaction):
    """
    Compare two Reaction objects to determine their relationship.
    
    Args:
        rxn1: The first Reaction object.
        rxn2: The second Reaction object.
    
    Returns:
        "SAME" if the reactions have identical substrates and products.
        "INVERTED" if one reaction is the reverse of the other.
        "DISTINCT" if the reactions are different.
    """
    rxn1Subs = collections.Counter(rxn1.Substrates)
    rxn1Prods = collections.Counter(rxn1.Products)
    rxn2Subs = collections.Counter(rxn2.Substrates)
    rxn2Prods = collections.Counter(rxn2.Products)

    if rxn1Subs == rxn2Subs and rxn1Prods == rxn2Prods:
        return "SAME"
    elif rxn1Subs == rxn2Prods and rxn1Prods == rxn2Subs:
        return "INVERTED"
    else:
        return "DISTINCT"



# Creates a chemical reaction network capable of storing reactions and components. Can run simulations once the network is properly assembled.
class ReactionNetwork:
    """
    Creates a ReactionNetwork object that contains all the reactions and their constituent components in a given chemical network. Can run 
    simulations of the network with optional network state persistence, and accepts modifications to reaction conditions in-between simulation runs.
    
    -- Functions --
    __init__ (Initializing function)
        No parameters required. Initializes an empty ReactionNetwork.

    addRxn:
        Adds a unique Reaction object to the ReactionNetwork.
    
    addRxns:
        Adds a list of Reaction objects to the ReactionNetwork.
    
    addComponent:
        Adds or overrides a Component object in the ReactionNetwork.
    
    addComponents:
        Adds a list of Component objects to the ReactionNetwork.
    
    disableModifying:
        Prevents modifications of the network. Useful when maintaining reaction network data structures.
    
    simulateReactionFn(simTime, continueFromLast=True, method="LSODA", simDataResolution=0, evalPts=[]):
        Simulates the reaction network over the specified time. Results are stored in SimResults.
    
    getCompiledCRNMatricies:
        Returns compiled matrices for ODE calculations.
    
    getCompiledReactionODEFunction:
        Returns the component name list and a derivative function for ODE integration.
    
    clearSimResults:
        Clears all simulation results.


    -- Properties --
    Reactions: Dictionary of Reaction objects in the network, keyed by their reaction strings.
    Components: Dictionary of Component objects in the network, keyed by their names.
    SimResults: List of simulation result dictionaries containing time-series data for each component.
    SimLastPoint: ComponentGroup containing the final concentrations from the last simulation.
    Modifiable: Indicates whether the network structure can still be modified.
    CompNameList: List of component names in the order used for ODE calculations.
    
    
    """
    def __init__(self):
        self._Reactions = {}
        self._Components = {}
        self._SimResults = []
        self._SimLastPoint = None
        self._Modifiable = True
        self._CompNameList = None


    @property
    def Reactions(self):
        return self._Reactions
    
    @property
    def Components(self):
        return self._Components
    
    @property
    def SimResults(self):
        return self._SimResults
    
    @property
    def SimLastPoint(self):
        return self._SimLastPoint
    
    @property
    def Modifiable(self):
        return self._Modifiable
    
    @property
    def CompNameList(self):
        return self._CompNameList


    # Checks to see if modifying the network is allowed, throws an error if not
    def _tryModifying(self):
        if not self._Modifiable:
            raise ValueError(
                "The reaction network structure has been locked, either from compiled " \
                "simulation data reliant on the current structure or due to external dependencies."
            )


    # Prevent modifications of the network. This is useful when using the CRN library to maintain reaction network data structures, as turning this on prevents the user from accidentally modifying the network and causing unintended behavior.
    def disableModifying(self):
        self._Modifiable = False

 
    #  Adds a unique Reaction object to ReactionNetwork.
    def addRxn(self, rxnObj: Reaction):
        self._tryModifying()
        for curRxn in self._Reactions.values():
            rxnRelation = _compareReactions(curRxn, rxnObj)
            if rxnRelation == "SAME":
                raise ValueError("This reaction already exists in the network.")
            elif rxnRelation == "INVERTED":
                raise ValueError("This reaction's inverse already exists in the network. Add a backward rate instead.")
        self._Reactions[rxnObj.RxnStr] = rxnObj
        for compName in set(rxnObj.Substrates+rxnObj.Products):
            if not compName in self._Components.keys():
                self.addComponent(Component(compName, 0.0), directFromRxn = True)


    # Adds a list of Reaction objects to ReactionNetwork.
    def addRxns(self, rxnObjList: list[Reaction]):
        for rxn in rxnObjList:
            self.addRxn(rxn)


    # Adds or overrides a Component object in the ReactionNetwork.
    def addComponent(self, componentObj: Component, directFromRxn: bool = False):
        self._tryModifying()
        if componentObj.Name == "Time":
            raise ValueError("'Time' is a protected name used to reference time stamps for a solution; please use a different component name.")
        elif not componentObj.Name in self._Components and not directFromRxn:
            warnings.warn(f"No reaction currently in the network consumes or produces the component '{componentObj.Name}'.", UserWarning)
        self._Components[componentObj.Name] = componentObj


    # Adds a list of Component objects to ReactionNetwork.
    def addComponents(self, compObjList: list[Component]):
        for comp in compObjList:
            self.addComponent(comp)


    def getCompiledCRNMatricies(self, pruneDisabledBackward: bool = True):
        compCount = len(self._Components)
        if not self._CompNameList:
            compNameList = list(self._Components.keys()) # Maintain a pre-set order of components when transferring the data into a linear vector for the ODE calculation, as to not mis-map output concentrations.
        else:
            compNameList = self._CompNameList

        compiledRateList = []
        compiledRxnMatrix = []
        compiledModifierMatrix = []

        def matrixAssembly(substrs, prods):
            subsList = [0]*(compCount)
            negSubsPosProdsList = [0]*(compCount)
            
            for substr in substrs:
                curInd = compNameList.index(substr)
                subsList[curInd] += 1
                negSubsPosProdsList[curInd] -= 1
            
            for prod in prods:
                curInd = compNameList.index(prod)
                negSubsPosProdsList[curInd] += 1
            
            compiledRxnMatrix.append(subsList)
            compiledModifierMatrix.append(negSubsPosProdsList)
        
        rxnNameList = [] # Pass along the list of reaction identifiers to keep track of the pre-set order
        for rxn in self._Reactions.values():
            compiledRateList.append(rxn.ForwardRate)
            rxnNameList.append((rxn.RxnStr, "ForwardRate"))
            matrixAssembly(rxn.Substrates, rxn.Products)

            if not pruneDisabledBackward or rxn.BackwardRate > 0:
                compiledRateList.append(rxn.BackwardRate)
                rxnNameList.append((rxn.RxnStr, "BackwardRate"))
                matrixAssembly(rxn.Products, rxn.Substrates)
        
        return compNameList, rxnNameList, compiledRateList, compiledRxnMatrix, compiledModifierMatrix
    
    
    def getCompiledReactionODEFunction(self):
        compNameList, _, compiledRateList, compiledRxnMatrix, compiledModifierMatrix = self.getCompiledCRNMatricies()
        compiledRateList = np.array(compiledRateList)
        compiledRxnMatrix = np.array(compiledRxnMatrix)
        compiledModifierMatrixTransposed = np.array(compiledModifierMatrix).T

        if np.max(compiledRxnMatrix) > 1: # Stoichiometries present > 1, so exponentiation is required
            def calculateDerivative(tVal, dVals):
                return np.matmul( # Distribute and sum the rates, with signs according to their status as a substrate or product of each reaction
                    compiledModifierMatrixTransposed,
                    np.multiply( # Multiply the kinetic rates by the np.prod result to obtain total rates for each reaction
                        compiledRateList,
                        np.prod( # Multiply concentrations of substrates consumed
                            np.power( # Distribute the substrates' concentrations into the sparse matrix format and exponentiate based on stoichiometry
                                dVals,
                                compiledRxnMatrix
                            ),
                            axis=1
                        )
                    )
                )
        else:
            compiledRxnMatrix = np.array(compiledRxnMatrix, dtype=bool)
            compiledRxnMatrixToggled = ~compiledRxnMatrix # Apply NOT filter

            def calculateDerivative(tVal, dVals):
                return np.matmul( # Distribute and sum the rates, with signs according to their status as a substrate or product of each reaction
                    compiledModifierMatrixTransposed,
                    np.multiply( # Multiply the kinetic rates by the np.prod result to obtain total rates for each reaction
                        compiledRateList,
                        np.prod( # Multiply concentrations of substrates consumed (all are to the power of one so we can use this trick to mimic np.power at a fraction of the compute cost)
                            compiledRxnMatrix * dVals + compiledRxnMatrixToggled,
                            axis=1
                        )
                    )
                )
            
        return compNameList, calculateDerivative
    

    # Simulates the reaction
    def simulateReactionFn(
        self,
        simTime: float,
        continueFromLast: bool = True,
        method: str = "LSODA",
        simDataResolution: int = 0,
        evalPts: list[float] = []
    ):
        """
        Run an ODE simulation of the reaction network.
        
        Integrates the system of ODEs defined by the reaction network using
        scipy's solve_ivp. Results are stored in SimResults and the final
        state is saved in SimLastPoint.
        
        Args:
            simTime: Total simulation time.
            continueFromLast: Chooses whether to continue from the last simulation's final state or from initial component concentrations.
            method: ODE solver method (default "LSODA" handles stiff systems well).
            simDataResolution: Number of evenly-spaced time points to record.
            evalPts: Explicit list of time points to record. Overrides simDataResolution.
        
        Note:
            Running a simulation locks the network structure (Modifiable becomes False).
        """
        if simDataResolution == 0 and len(evalPts) == 0:
            raise ValueError("'evalPts' must have at least one time entry or 'simDataresolution' must be > 0 to record simulation results.")
        elif len(evalPts) == 0:
            evalPts = np.linspace(0, simTime, simDataResolution)
        else:
            for x in evalPts:
                if x > simTime:
                    raise ValueError("All evaluated time points in 'evalPts' must be within the designated simulation time.")

        compNameList, reactionODEFn = self.getCompiledReactionODEFunction()
        if self._Modifiable:
            self.disableModifying()
        if not self._CompNameList:
            self._CompNameList = compNameList
        
        freshStart = False
        if not continueFromLast or not self._SimLastPoint:
            freshStart = True
            self._SimLastPoint = ComponentGroup({comp.Name: comp.Concentration for comp in self._Components.values()})

        rawSolution = spi.solve_ivp( # Run scipy's solve_ivp function to integrate the ODE system over time. It is recommended that the method chosen should be capable of handling stiff ODE systems (default should handle most cases)
            reactionODEFn,
            (0, simTime),
            [self._SimLastPoint._Components[compName] for compName in compNameList],
            method=method,
            t_eval=evalPts
        )

        self._SimResults.append({compNameList[i]: x for i,x in enumerate(rawSolution.y)})
        self._SimResults[-1]["Time"] = rawSolution.t + (0 if freshStart else self._SimResults[-2]["Time"][-1]) # If continuing from a prior simulation, increase the Time values based on previous simulation's last time point.
        self._SimLastPoint = ComponentGroup({compName: self._SimResults[-1][compName][-1] for compName in compNameList})


    # Allows a simulation result to have a modification made to the last simulation's concentration profile.
    # Useful if you want to simulate spiking a component in a solution (simulating a sudden manual/external adjustment in concentration)
    def adjustResultComponent(self, compName: str, compConc: float):
        if not self._SimLastPoint:
            raise ValueError("A prior simulation has not yet been recorded.")
        self._SimLastPoint._Components[compName] = compConc

    def clearSimResults(self, regainModifiability: bool = True):
        self._SimLastPoint = None
        self._SimResults = []
        self._Modifiable = regainModifiability
        self._CompNameList = None


# Implicitly creates a string of contents of component concentrations in the correct formatting while also maintaining the desired data structure.
class ComponentGroup:
    """
    Implicitly creates a string of contents of component concentrations in the correct formatting 
    while also maintaining the desired data structure. Used to store and display simulation state.

    -- Functions --
    __init__ (Initializing function)
        componentConcDict:
            Dictionary mapping component names (strings) to their concentrations (numbers).


    -- Properties --
    Components: Internal dictionary storing the component-concentration mapping.
    
    """
    def __init__(self, componentConcDict: dict[str, Component]):
        self._Components = componentConcDict

    def __str__(self):
        return "\n".join([
            f"{itm} = {conc:.4g}"
            for itm, conc in self._Components.items()
        ])


############### Behavior Manager ##############
#                                             #
#               By: Colin Yancey              #
#            Created September 2025           #
#            Last Edited: 02/24/2026          #
#                  Schulman Lab               #
#            Johns Hopkins University         #
#                                             #
###############################################

import jax
jax.config.update("jax_enable_x64", True)
import jax.numpy as jnp
import warnings

from .behavior import BehaviorTimeCourse
from ..core.crnData import (
    CRNInfoHandler
)
from ..core.simulate import (
    EvaluationPointsHandler,
    CRNSimulationRunner
)
from ..optimization.optimize import (
    CRNOptimizationFramework
)
from ..utils.helpers import (
    ManagerClass
)

###############################################


class OptimizeBehaviorsManager(ManagerClass):
    def __init__(self, crnInfo: CRNInfoHandler, behaviorTimeCourses: list[BehaviorTimeCourse], highMin:float = 0.8, boundMult:float = 8, stochasticity = 0.02):
        self.CRNInfo = crnInfo
        self.BehaviorTimeCourses = behaviorTimeCourses

        self.NormalizedWeights = jnp.array([bTC.Weight for bTC in behaviorTimeCourses])
        self.NormalizedWeights /= jnp.sum(self.NormalizedWeights)

        self.EvaluationPointsHandlers = [EvaluationPointsHandler(behaviorTimeCourseObj.UniqueEvaluationPoints) for behaviorTimeCourseObj in self.BehaviorTimeCourses]

        # If true that all resolutions match point counts, then no specialized point matching necessary
        self.NeedsStochasticity = not all(
            behavior.PointCount == behavior.Resolution
            for bTC in self.BehaviorTimeCourses
            for bG in bTC.BehaviorGroups
            for behavior in bG.Behaviors
        )
        if self.NeedsStochasticity:
            warnings.warn(
                f"A behavior has been detected to have a point count different from its resolution (likely from providing requiredMatch).\n" \
                f"  Using a stochasticity of {stochasticity}. The higher the stochasticity, the less the bias toward cost function evaluations that performed well.",
                UserWarning
            )
            self.Stochasticity = stochasticity
        
        if highMin < 0.2 or highMin > 0.99:
            raise ValueError("The minimum HIGH threshold must be set to a value between 0.2 and 0.99.")
        if boundMult < 2:
            raise ValueError("Bound multiple must be set to a value >= 2.")
        
        # Memoize all the precursor math for attached functions
        self._HighMin = highMin
        self._LowMax = self._HighMin/boundMult
        
        self._CoefLow = 2*(self._LowMax - 1)
        self._DisplaceLow = self._LowMax - self._CoefLow*jnp.log(1-self._LowMax)

        self._DivHigh = 1 - self._HighMin
        self._CoefHigh = 2*self._HighMin
        self._DisplaceHigh = self._DivHigh + self._CoefHigh*jnp.log(self._HighMin)
    
    def getHighMin(self):
        return self._HighMin

    def getLowMax(self):
        return self._LowMax
    
    def getBFLow(self, x):
        return jnp.where(
            x <= self._LowMax,
            x**2/self._LowMax,
            jnp.where(
                x < 1,
                self._CoefLow*jnp.log(1 - x) + self._DisplaceLow,
                1e6  # This value (1.0) should never be reached under normal use cases.
            )
        )
    
    def getBFHigh(self, x):
        return jnp.where(
            x <= 0,
            1e6,  # This value (0.0) should never be reached under normal use cases.
            jnp.where(
                x < self._HighMin,
                -self._CoefHigh*jnp.log(x) + self._DisplaceHigh,
                (1 - x)**2/self._DivHigh
            )
        )
    
    def getMinimumSum(self, pointCosts, pointCount: int):
        resolution = pointCosts.shape[0]
        selectionLogits = -pointCosts / self.Stochasticity # More stochasticity = flatter cost landscape, cost incorporates more of an average point blend but you get a smoother derivative landscape.

        def selectionStep(_, loopState):
            currentRemainingSelectionMass, currentMinimumCostSum = loopState
            maskedLogits = selectionLogits + jnp.log(currentRemainingSelectionMass + 1e-12)

            # Differentiable "pick one more low-cost point" distribution.
            selectionProbabilities = jax.nn.softmax(maskedLogits)

            # Add expected selected cost for this step.
            updatedMinimumCostSum = currentMinimumCostSum + jnp.sum(selectionProbabilities * pointCosts)

            # Reduce future probability mass for points already selected now.
            updatedRemainingSelectionMass = currentRemainingSelectionMass * (1.0 - selectionProbabilities)

            return updatedRemainingSelectionMass, updatedMinimumCostSum

        remainingSelectionMass = jnp.ones(resolution)
        minimumCostSum = jnp.array(0.0, dtype=pointCosts.dtype)

        # Rolls up a for loop that iterates over point counts to 
        _, minimumCostSum = jax.lax.fori_loop(0, pointCount, selectionStep, (remainingSelectionMass, minimumCostSum))
        return minimumCostSum / pointCount

    def getCostFn(self, optimizerFrameworkObj: CRNOptimizationFramework, simulationRunnerObj: CRNSimulationRunner):
        crnInfo = self.CRNInfo
        bfLow = self.getBFLow
        bfHigh = self.getBFHigh
        normedWeights = self.NormalizedWeights
        behaviorTimeCourses = self.BehaviorTimeCourses

        precomputedTimeCourseData = []
        for i, behaviorTimeCourseObj in enumerate(self.BehaviorTimeCourses):
            timeDomainData = self.EvaluationPointsHandlers[i].parseEvalPts(behaviorTimeCourseObj.TimeCourse)
            precomputedTimeCourseData.append(timeDomainData)

        if self.NeedsStochasticity:
            def getBehaviorGroupCost(behaviorGroup, resultOfInterest):
                groupCost = 0
                startInd = 0
                for behavior in behaviorGroup.Behaviors:
                    endInd = startInd + behavior.Resolution
                    segmentedCosts = jax.vmap(bfHigh if behavior.BehaviorType == "HIGH" else bfLow)(resultOfInterest[startInd : endInd])
                    groupCost += self.getMinimumSum(segmentedCosts, behavior.PointCount) * behavior.Weight
                    startInd = endInd

                return groupCost / sum(b.Weight for b in behaviorGroup.Behaviors)
        else:
            def getBehaviorGroupCost(behaviorGroup, resultOfInterest):
                lowCosts = jax.vmap(bfLow)(resultOfInterest)
                highCosts = jax.vmap(bfHigh)(resultOfInterest)
                return jnp.sum(
                    jnp.dot(
                        jnp.where(
                            behaviorGroup.EvaluationIsHigh,
                            highCosts,
                            lowCosts
                        ),
                        behaviorGroup.EvaluationPointWeights
                    )
                )
        
        
        def getAllBehaviorCost(y):
            initCondInfo = optimizerFrameworkObj.getStartingConditions(y, crnInfo)

            behaveCost = 0
            for i, behaviorTimeCourseObj in enumerate(behaviorTimeCourses):
                results = simulationRunnerObj.runTimeCourse(precomputedTimeCourseData[i], crnInfo, initCondInfo)
                courseCost = 0

                for j, mappingGroup in enumerate(behaviorTimeCourseObj.EvaluationPointMappingGroups):
                    behaviorGroup = behaviorTimeCourseObj.BehaviorGroups[j]
                    
                    resultOfInterest = behaviorGroup.getNormalizeVMAP(initCondInfo.ComponentsDictionary)(
                        jnp.sum(
                            jnp.atleast_2d(
                                results[mappingGroup, behaviorGroup.getRelevantComponentIndices(crnInfo.CompNameList)]
                            ),
                            axis = 0
                        )
                    )
                    courseCost += getBehaviorGroupCost(behaviorGroup, resultOfInterest) * behaviorTimeCourseObj.BehaviorGroupWeights[j]
                    
                behaveCost += courseCost * normedWeights[i]
            return behaveCost
        
        return getAllBehaviorCost


############### Time Course Info ##############
#                                             #
#               By: Colin Yancey              #
#            Created September 2025           #
#            Last Edited: 02/24/2026          #
#                  Schulman Lab               #
#            Johns Hopkins University         #
#                                             #
###############################################


from dataclasses import dataclass, field
from typing import Optional, Literal, Callable

from ..crn import (
    Component,
    Reaction
)
from ..utils.constants import (
    ALLOWED_INTERRUPTIONS
)
from ..utils.base import (
    DynamicFunctionStoring,
    
    _modifyBySubstitute,
    _modifyByAdd,
    _modifyBySubtract
)
from ..utils.base import (
    _getValueName,
    _getParameterName
)

###############################################

# At time breaks, the reaction conditions of a CRN may change. Interruption objects represent these changes.
class Interruption(DynamicFunctionStoring):
    def __init__(
        self,
        interruptionObject: Component | Reaction,
        valueObtain: int | float | Callable[[dict, dict], float | int],
        valueName: Optional[Literal["ForwardRate", "BackwardRate", "Concentration"]] = None,
        interruptionType: Literal["SUBSTITUTE", "ADD", "SUBTRACT"] = "SUBSTITUTE"
    ):
        self.ValueName = _getValueName(interruptionObject, valueName)
        self.Name = _getParameterName(interruptionObject)
        self.InterruptionType = interruptionType

        if self.InterruptionType.upper() not in ALLOWED_INTERRUPTIONS:
            raise ValueError(f"InterruptionType must be one of {ALLOWED_INTERRUPTIONS}")
        
        if callable(valueObtain):
            self.getValue = valueObtain
        else:
            self.DefaultValue = valueObtain
    
        if interruptionType == "SUBSTITUTE":
            self.getModifiedList = _modifyBySubstitute
        elif interruptionType == "ADD":
            self.getModifiedList = _modifyByAdd
        elif interruptionType == "SUBTRACT":
            self.getModifiedList = _modifyBySubtract

    def getValue(self, _, __):
        return self.DefaultValue


# Represents a period of time where a CRN runs independent of reaction condition changes or sudden concentration
# jumps from external manipulation. A list of TimeSpan objects represents a full time course for a simulation
# run by the CRNOptimizer class. TimeSpan objects can have a list of Interruption objects that represent
# the changes made to the conditions of the experiment in between TimeSpan intervals.
@dataclass
class TimeSpan:
    Time: float
    Interruptions: list[Interruption] = field(default_factory=list)



################ Initialization ###############
#                                             #
#               By: Colin Yancey              #
#            Created September 2025           #
#            Last Edited: 02/24/2026          #
#                  Schulman Lab               #
#            Johns Hopkins University         #
#                                             #
###############################################

import jax
jax.config.update("jax_enable_x64", True) # Maintain 64-bit precision.
import jax.numpy as jnp

import diffrax

from .timeCourse import (
    TimeSpan
)
from .crnData import (
    CRNInfoHandler,
    InitialConditionInfo
)

###############################################

class EvaluationPointsHandler:
    def __init__(self, totalEvalPts: jnp.ndarray):
        self.TotalEvalPts = totalEvalPts
        self.ParsedTimeCourseData = {}
    
    def parseEvalPts(self, timeCourse: list[TimeSpan]):
        timeCourseKey = id(timeCourse)
        if not timeCourseKey in self.ParsedTimeCourseData:
            currentTime = 0
            timeCourseData = []

            for i, timeSpan in enumerate(timeCourse):
                if i+1 != len(timeCourse):
                    spanMask = (self.TotalEvalPts >= currentTime) & (self.TotalEvalPts < currentTime + timeSpan.Time)
                else:
                    spanMask = (self.TotalEvalPts >= currentTime) & (self.TotalEvalPts <= currentTime + timeSpan.Time) # Account for the final point if this is the final TimeSpan
                spanIndices = jnp.where(spanMask)[0] # Get list of relevant indices
                evalPts = jnp.zeros(len(spanIndices) + 1)
                evalPts = evalPts.at[-1].set(timeSpan.Time) # Make sure the last time is being saved regardless, as this point is the initial condition vector for the next timespan
                if len(spanIndices) > 0: # Put in the rest of the span index values
                    evalPts = evalPts.at[:-1].set(self.TotalEvalPts[spanIndices] - currentTime)
                
                timeCourseData.append((timeSpan, spanIndices, evalPts))
                currentTime += timeSpan.Time
            
            self.ParsedTimeCourseData[timeCourseKey] = timeCourseData
            
        return self.ParsedTimeCourseData[timeCourseKey]
    

class CRNSimulationRunner:
    def __init__(self, rateMinimum: float = 0, concentrationMinimum: float = 0, dt0Min: float = 1e-3, dtmin: float = 1e-15, rtol: float = 1e-4, atol: float = 1e-7):
        self.RateMinimum = rateMinimum
        self.ConcentrationMinimum = concentrationMinimum

        self.dt0Min = dt0Min
        self.dtmin = dtmin
        self.rtol = rtol
        self.atol = atol

    def runTimeCourse(self, timeCourseData, crnInfo: CRNInfoHandler, initCondInfo: InitialConditionInfo, t0Time = 0):
        allResults = []

        concList = initCondInfo.ConcentrationList
        rateList = initCondInfo.RateList

        def odeFn(t, y, args):
            rates = rateList * jnp.prod(jnp.maximum(y, self.ConcentrationMinimum) ** crnInfo.ReactionMatrix, axis=1)
            return jnp.dot(crnInfo.ModifierMatrixTransposed, rates)

        for timeSpan, spanIndices, evalPts in timeCourseData:
            for mod in timeSpan.Interruptions:
                newVal = mod.getValue(initCondInfo.ReactionsDictionary, initCondInfo.ComponentsDictionary)
                if mod.ValueName == "Concentration":
                    concList = mod.getModifiedList(concList, crnInfo.CompNameList.index(mod.Name), newVal, self.ConcentrationMinimum)
                else:
                    rateList = mod.getModifiedList(rateList, crnInfo.RxnNameList.index((mod.Name, mod.ValueName)), newVal, self.RateMinimum)

            try:
                solution = diffrax.diffeqsolve(
                    diffrax.ODETerm(odeFn),
                    solver = diffrax.Kvaerno5(),
                    t0 = t0Time,
                    t1 = timeSpan.Time,
                    dt0 = jnp.minimum(timeSpan.Time / 10000, self.dt0Min),
                    y0 = concList,
                    saveat = diffrax.SaveAt(ts = evalPts),
                    stepsize_controller = diffrax.PIDController(
                        rtol = self.rtol,
                        atol = self.atol,
                        dtmin = self.dtmin,
                        dtmax = timeSpan.Time/100,
                        safety = 0.9
                    ),
                    max_steps = 50000
                )
            except Exception as e:
                raise ValueError(f"Integration failure during {timeSpan}:\n {e}")
            
            if len(spanIndices) > 0:
                allResults.append(solution.ys[:-1, :])
            concList = solution.ys[-1, :]
        
        return jnp.concatenate(allResults)



####### Parameter Optimization Visualizer #####
#                                             #
#               By: Colin Yancey              #
#             Created January 2026            #
#            Last Edited: 02/24/2026          #
#                  Schulman Lab               #
#            Johns Hopkins University         #
#                                             #
###############################################

from pathlib import Path
from typing import Optional
from numpy.typing import NDArray
import numpy as np
import jax.numpy as jnp

import imageio.v2 as iio
import os

import matplotlib.pyplot as plt

from ..results.optimizerResults import CompiledOptimizationData
from .plotting import *

###############################################

def saveFreeParametersVisualization(
    compiledOptDataObject: CompiledOptimizationData,
    iterationList: Optional[list[int]] = None, # If provided, plot will only contain the selected iterations in the list, each denoted by either a number indicating the time course number, or a tuple containing the time course number and the behavior group number within that time course.
    fileLocation: Path = Path('./'),
    fileName: str = 'freeParameters',
    fps: int = 10,
    keepIndividualImages: bool = False,
    lossLineAccentColor: NDArray = np.array([0.122, 0.467, 0.706]), # Coloring the prior iterations in the loss line plot
    baseColor: NDArray = np.array([0.75, 0.75, 0.75]), # Baseline bar color (default is neutral gray)
    posTargetColor: NDArray = np.array([0.5, 0.95, 0.5]), # Bar color when parameter is increasing rapidly (default is green)
    negTargetColor: NDArray = np.array([0.95, 0.5, 0.5]) # Bar color when parameter is decreasing rapidly (default is red)
):
    optRes = compiledOptDataObject.OptimizerRunResultObject
    paramHist = optRes.ParameterHistory
    totalIter = optRes.LastIteration
    parameters = optRes.CRNOptimizationFrameworkObject.Parameters
    paramNames = [str(param) for param in parameters]

    if iterationList is None:
        iterationList = np.arange(totalIter)
    else:
        iterationList = np.array(iterationList) - 1

    workspaceName = fileLocation / 'mp4 Output'
    os.makedirs(workspaceName, exist_ok=True)

    lossHist = optRes.LossHistory
    minLoss = jnp.min(jnp.array(lossHist)).item()
    maxLoss = jnp.max(jnp.array(lossHist)).item()
    multiplierLoss = (maxLoss / minLoss)**(1 / 20)
    xLoss = np.arange(totalIter)

    framePaths = []
    minValue = jnp.min(jnp.stack(paramHist)).item()
    maxValue = jnp.max(jnp.stack(paramHist)).item()
    multiplier = (maxValue / minValue)**(1 / 20)

    for iteration in iterationList:
        params = paramHist[iteration]

        fig, (ax_loss, ax) = plt.subplots(
            2, 1, figsize=(12, 8),
            gridspec_kw={'height_ratios': [1, 4]}
        )

        for a in (ax_loss, ax):
            a.spines['top'].set_visible(False)
            a.spines['right'].set_visible(False)
            a.grid(True, which='major', axis='y', linestyle='--', color='0.85', alpha=0.5)
            a.grid(True, which='minor', axis='y', linestyle=':', color='0.85', alpha=0.35)
            a.set_axisbelow(True)

        ax_loss.plot(xLoss[:iteration + 1], lossHist[:iteration + 1], color=lossLineAccentColor, linewidth=2.0) # Prior loss
        ax_loss.plot(xLoss[iteration:], lossHist[iteration:], color='0.6', linestyle='--', linewidth=1.5) # Future loss
        #ax_loss.fill_between(xLoss[:iteration + 1], lossHist[:iteration + 1], color=accentColor, alpha=0.08, step=None)
        ax_loss.plot(xLoss[iteration], lossHist[iteration], marker='o', markersize=5, markerfacecolor='white', markeredgecolor=lossLineAccentColor, markeredgewidth=1.5, zorder=5) # Add marker for current iteration

        ax_loss.set_ylabel('Loss', fontsize=11)
        ax_loss.set_xticks([])

        # Top: Loss history plot (log space)
        ax_loss.set_ylim(minLoss / multiplierLoss, maxLoss * multiplierLoss)
        ax_loss.set_yscale('log')
        ax_loss.set_axisbelow(True)
        ax_loss.set_xlim(0, totalIter - 1)

        ax_loss.set_ylabel('Loss')
        ax_loss.set_xticks([])  # keep top plot clean

        bars = ax.bar(np.arange(len(params)), params, tick_label = paramNames, color = baseColor, edgecolor = '0.25', linewidth = 1.5)
        for i, bar in enumerate(bars):
            x = bar.get_x()
            w = bar.get_width()
            ax.hlines(y=paramHist[0][i], xmin=x + w/3, xmax=x + 2*w/3, colors='0.25', linewidth=2.0, zorder=3) # Slightly thicker than the bar edge to distinguish it

        if iteration > 0:
            pct_change = (paramHist[iteration] - paramHist[iteration - 1]) / paramHist[iteration - 1]
            for i, bar in enumerate(bars):
                pc = pct_change[i]
                if pc >= 0:
                    finalColor = baseColor + (posTargetColor - baseColor) * min(pc, 0.25) / 0.25 # map [0, 0.25] -> [0, 1]
                else:
                    finalColor = baseColor + (negTargetColor - baseColor) * -max(pc, -0.2) / 0.2 # map [-0.2, 0] -> [1, 0]
                bar.set_facecolor(tuple(float(x) for x in finalColor))

        ax.set_xlabel('Free Parameters', fontsize=11)
        ax.set_ylabel('Parameter Value', fontsize=11)
        ax.tick_params(axis='y', labelsize=10)
        ax.set_xticklabels(paramNames, rotation=30, ha='right', fontsize=9)

        ax.set_ylim(minValue / multiplier, maxValue * multiplier)
        ax.set_yscale('log')
        ax.set_axisbelow(True)

        ax.grid(True, which='both', axis='y', linestyle='--', alpha=0.35)

        addIterationTracker(fig, iteration + 1, totalIter)

        fig.tight_layout()
        framePath = workspaceName / f"{fileName}_{iteration}.png"
        plt.savefig(framePath, dpi=100)
        plt.close(fig)
        framePaths.append(framePath)

    with iio.get_writer(workspaceName / f"{fileName}.mp4", fps = fps, codec = 'libx264', format = 'FFMPEG', ffmpeg_params = ['-pix_fmt', 'yuv420p']) as writer:
        for framePath in framePaths:
            image = iio.imread(framePath)
            writer.append_data(image)

    if not keepIndividualImages:
        for framePath in framePaths:
            try:
                os.remove(framePath)
            except OSError:
                pass



########### Iteration Parsing Handler #########
#                                             #
#               By: Colin Yancey              #
#             Created January 2026            #
#            Last Edited: 02/24/2026          #
#                  Schulman Lab               #
#            Johns Hopkins University         #
#                                             #
###############################################


###############################################


# Returns a list of iterations, having a selection frequency based on the size of the intermediate count and the iteration count.
# def getIterationList(intermediateCount = ,)


############### Experiment Data ###############
#                                             #
#               By: Colin Yancey              #
#            Created September 2025           #
#            Last Edited: 02/24/2026          #
#                  Schulman Lab               #
#            Johns Hopkins University         #
#                                             #
###############################################

import jax.numpy as jnp
from scipy.signal import savgol_filter

from typing import Callable
import warnings

from ..core.crnData import (
    CRNInfoHandler
)
from ..core.simulate import (
    TimeSpan,
    EvaluationPointsHandler
)
from ..crn import (
    ReactionNetwork
)
from ..utils.base import (
    ComponentCompositesHandler,
    DynamicFunctionStoring,
    DataStore
)

###############################################

# Stores data points that represent the fraction of maximum relevant fluorescence within an experiment alongside the model for obtaining this fluorescence from the CRN results.
class Signal(ComponentCompositesHandler, DynamicFunctionStoring):
    def __init__(
        self,
        compNames: str | list[str],
        maxSignalConc: float | int | Callable[[dict], float],
        dataPoints: jnp.ndarray,
        weight: float = 1.0, # Must be >= 0 (although, why bother including it if it is 0?)
        gradientImpactRadius: int = 5, # Must be >= 0
        gradientImpactSigma: float = 2.0, # Must be > 0
        compositesAllowed: bool = False
    ):
        ComponentCompositesHandler.__init__(self, compNames, compositesAllowed)

        if callable(maxSignalConc):
            self.getMaxSignal = maxSignalConc
        else:
            self.MaxSignal = maxSignalConc
        
        self.DataPoints = dataPoints

        self.Weight = weight
        # Use absolute gradient magnitude so rising/falling regions both increase impact.
        weights = jnp.abs(
            savgol_filter(
                self.DataPoints,
                window_length = gradientImpactRadius,
                polyorder = 1,
                deriv=1
            )
        )
        totalWeight = jnp.sum(weights)
        
        if totalWeight > 0:
            if gradientImpactRadius > 0:
                # Spreads any strong signal gradients to nearby points so "near-gradient" regions also retain impactful weights.
                offsets = jnp.arange(-gradientImpactRadius, gradientImpactRadius + 1)
                kernel = jnp.exp(-0.5 * jnp.square(offsets / gradientImpactSigma))
                kernel /= jnp.sum(kernel)
                weights = jnp.convolve(weights, kernel, mode="same")

            totalWeight = jnp.sum(weights)
            self.GradientImpactVector = weights / totalWeight
        else:
            self.GradientImpactVector = jnp.ones_like(weights) / weights.size
    
    def getMaxSignal(self, _):
        return self.MaxSignal
    
    def normalizeValue(self, valToNorm, concs):
        return valToNorm / self.getMaxSignal(concs)
        


# Stores a given experiment's progression through the time course and defines what signals were collected.
class Experiment:
    def __init__(self, timeCourse: list[TimeSpan], signals: list[Signal], weight: float = 1.0):
        self.TimeCourse = timeCourse

        if len(signals) == 0:
            raise ValueError("An Experiment object must have at least one Signal object.")
        standardDPNum = signals[0]
        for signal in signals[1:]:
            if len(signal.DataPoints) != standardDPNum:
                raise ValueError("All Signal objects in an Experiment object must have the same number of data points.")
        self.Signals = signals
        self.Weight = weight

        self.SignalWeights = jnp.array([signal.Weight * signal.GradientImpactVector for signal in self.Signals])
        self.SignalWeights /= jnp.sum(jnp.array([signal.Weight for signal in self.Signals]))


# Stores experiments that were run simultaneously, thus sharing data recording times and ideally CRN charactertistics. Differences in CRN initial conditions are handled by
# each individual experiment's time course interruptions.
class ExperimentGroup(DataStore):
    def __init__(self, timePointsHandler: EvaluationPointsHandler, crn: ReactionNetwork, experiments: list[Experiment], weight: float = 1.0):
        self.TimePointsHandler = timePointsHandler
        self.CRNInfo = CRNInfoHandler(crn)
        self.Experiments = experiments
        self.Weight = weight

        standardDPNum = len(self.TimePointsHandler.TotalEvalPts)
        for experiment in self.Experiments:
            for signal in experiment.Signals:
                if len(signal.DataPoints) != standardDPNum:
                    raise ValueError(
                        "All Experiment objects in an ExperimentGroup object must have the same number of data points in each of their Signal " \
                        "objects as the given list of associated times in the EvaluationPointsHandler."
                    )

        self.ExperimentWeights = jnp.array([expt.Weight for expt in self.Experiments])
        self.ExperimentWeights /= jnp.sum(self.ExperimentWeights)*len(self.TimePointsHandler.TotalEvalPts)
    

    # This class specifically has a weight adjustment to accomodate the cumulative data curation process. Weights should generally be decided on instance generation for maximum safety.
    def adjustWeight(self, weight: float):
        warnings.warn(f"ExperimentGroup weight adjustments should be made before incorporating them into an OptimizeExperimentsManager instance.", UserWarning)
        self.Weight = weight



################ Initialization ###############
#                                             #
#               By: Colin Yancey              #
#            Created September 2025           #
#            Last Edited: 02/24/2026          #
#                  Schulman Lab               #
#            Johns Hopkins University         #
#                                             #
###############################################

from .constants import (
    UNIT_WEIGHTS,
    
    ALLOWED_BEHAVIORS,
    ALLOWED_INTERRUPTIONS
)

from .base import (
    DEFAULT_VALUE_NAMES,

    DynamicFunctionStoring,
    ComponentCompositesHandler,
    DataStore,
    
    # Internal helper functions
    _getMinSecStr,
    getComponentHosts,
    _modifyBySubstitute,
    _modifyByAdd,
    _modifyBySubtract,
    _getValueName,
    _getParameterName
)

__all__ = [
    # Constants
    "UNIT_WEIGHTS",
    "ALLOWED_BEHAVIORS",
    "ALLOWED_INTERRUPTIONS",

    # Constants (with dependencies)
    "DEFAULT_VALUE_NAMES",
    
    # Helper Classes
    "DynamicFunctionStoring",
    "ComponentCompositesHandler",
    "DataStore",

    # Helper Functions
    "_getMinSecStr",
    "getComponentHosts",
    "_getValueName",
    "_getParameterName",
    "_modifyBySubstitute",
    "_modifyByAdd",
    "_modifyBySubtract"
]


########### Helper Functions/Classes ##########
#                                             #
#               By: Colin Yancey              #
#            Created September 2025           #
#            Last Edited: 03/03/2026          #
#                  Schulman Lab               #
#            Johns Hopkins University         #
#                                             #
###############################################

from abc import ABC, abstractmethod

from ..core.simulate import (
    CRNSimulationRunner
)
from ..optimization.optimize import (
    CRNOptimizationFramework
)

###############################################

class ManagerClass(ABC):
    @abstractmethod
    def getCostFn(self, optimizerFrameworkObj: CRNOptimizationFramework, simulationRunnerObj: CRNSimulationRunner):
        pass


# Used to debug visualization code errors where the wrong manager class is provided for a given visualizer.
def _checkManagerType(managerObj: ManagerClass, desiredClass: ManagerClass):
    if not isinstance(managerObj, desiredClass):
        raise TypeError(f"Expected the following manager class: '{desiredClass.__name__}'\n  Got class '{type(managerObj).__name__}'")



######### Base Helper Functions/Classes #######
#                                             #
#               By: Colin Yancey              #
#            Created September 2025           #
#            Last Edited: 03/03/2026          #
#                  Schulman Lab               #
#            Johns Hopkins University         #
#                                             #
###############################################

from typing import Literal
import time

import jax
jax.config.update("jax_enable_x64", True)  # Maintain 64-bit precision.
import jax.numpy as jnp

import inspect
import re

from pathlib import Path
import dill

from ..crn import (
    Component,
    Reaction
)

###############################################

# The default property that is chosen when creating a parameter for a given class.
DEFAULT_VALUE_NAMES = {
    Component.__name__: "Concentration",
    Reaction.__name__: "ForwardRate"
}


def getComponentHosts(compName: str, compCandidates: list[str], delimiter: str = ":"):
    delimEscape = re.escape(delimiter)
    # Start of string/delimiter -> compName <- End of string/a delimiter
    pattern = f"(^|{delimEscape}){re.escape(compName)}($|{delimEscape})"
    
    return [s for s in compCandidates if re.search(pattern, s)]


def _getMinSecStr(timeVal: int | float):
    return f"{int(timeVal // 60)} min {(timeVal % 60):.3f} sec"


def _getValueName(parameterObject: Component | Reaction, valueName: Literal["ForwardRate", "BackwardRate", "Concentration", None]):
    paramType = type(parameterObject).__name__
    if not valueName:
        return DEFAULT_VALUE_NAMES[paramType]
    if isinstance(parameterObject, Component):
        if not valueName == "Concentration":
            print(f"Warning: Provided '{valueName}' as value name for a Component object. Reverted to '{DEFAULT_VALUE_NAMES[Component]}'.")
        return DEFAULT_VALUE_NAMES[Component]
    elif not valueName in ["ForwardRate", "BackwardRate"]:
        print(f"Warning: Provided '{valueName}' as a value name for a Reaction object. Reverted to '{DEFAULT_VALUE_NAMES[Reaction]}'.")
        return DEFAULT_VALUE_NAMES[Reaction]
    return valueName


def _getParameterName(parameterObject):
    if isinstance(parameterObject, Component):
        return parameterObject.Name
    else:
        return parameterObject.RxnStr


def _modifyBySubstitute(listOfInterest, newInd, newVal, modValMin):
    return listOfInterest.at[newInd].set(jnp.maximum(modValMin, newVal))

def _modifyByAdd(listOfInterest, newInd, newVal, modValMin):
    return listOfInterest.at[newInd].set(jnp.maximum(modValMin, listOfInterest[newInd] + newVal))

def _modifyBySubtract(listOfInterest, newInd, newVal, modValMin):
    return listOfInterest.at[newInd].set(jnp.maximum(modValMin, listOfInterest[newInd] - newVal))


class DynamicFunctionStoring:
    def __getstate__(self):
        state = self.__dict__.copy()
        state["_sourceFuncData"] = []
        for name, val in inspect.getmembers(self):
            if callable(val):
                if isinstance(val, type(self.__dict__.get(name))):
                    # Exception for classes, since these are handled by dill properly.
                    if not (inspect.isfunction(val) or inspect.ismethod(val)):
                        continue
                    valSourceCode = inspect.getsource(val)
                    freeVarData = {}
                    if val.__closure__ is not None:
                        freeVarData = {
                            freeVarName: val.__closure__[i].cell_contents
                            for i, freeVarName in enumerate(val.__code__.co_freevars)
                        }
                    state["_sourceFuncData"].append((valSourceCode.strip(), freeVarData))

        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        for funcData in self._sourceFuncData:
            funcSource, freeVarData = funcData
            for freeVarName, freeVarVal in freeVarData.items():
                self.__dict__[freeVarName] = freeVarVal

            exec(funcSource, self.__dict__)



class ComponentCompositesHandler:
    """
    Base class for handling component composites in concentration lookups.
    
    Subclasses or instances should define `normalizeValue(normVal, compDict)` 
    either by:
    - Assigning a function to `self.normalizeValue` in `__init__`
    - Overriding the method in a subclass
    """
    def __init__(self, compNames: str | list[str], compositesAllowed: bool):
        if isinstance(compNames, str):
            self.ComponentNames = [compNames]
        else:
            self.ComponentNames = compNames
        self.CompositesAllowed = compositesAllowed

        self.RelevantComponentIndices = {}

    
    def getRelevantComponentIndices(self, allCompNames):
        allCompNameKey = id(allCompNames)
        if not allCompNameKey in self.RelevantComponentIndices:
            compIndices = []
            if self.CompositesAllowed:
                for baseCompName in self.ComponentNames:
                    for compName in getComponentHosts(baseCompName, allCompNames):
                        compIndices.append(allCompNames.index(compName))
            else:
                for compName in self.ComponentNames:
                    compIndices.append(allCompNames.index(compName))
            
            self.RelevantComponentIndices[allCompNameKey] = compIndices
            
        return self.RelevantComponentIndices[allCompNameKey]


    def getNormalizeVMAP(self, compDict):
        return jax.vmap(lambda normVal: self.normalizeValue(normVal, compDict))



class DataStore:
    def save(self, fileLocation: Path, fileName: str):
        self.SavedTimestamp = time.time()
        fullResName = fileName + ".pkl"
        with open(fileLocation / fullResName, 'wb') as f:
            dill.dump(
                self,
                f
            )
    
    @classmethod
    def load(cls, filePath: Path):
        with open(filePath, 'rb') as f:
            data = dill.load(f)
        if not isinstance(data, cls):
            print(f"Expected instance of {cls.__name__} or subclass; got {type(data).__name__}.")
        return data



################# Behavior Data ###############
#                                             #
#               By: Colin Yancey              #
#            Created September 2025           #
#            Last Edited: 02/24/2026          #
#                  Schulman Lab               #
#            Johns Hopkins University         #
#                                             #
###############################################

import jax
jax.config.update("jax_enable_x64", True)
import jax.numpy as jnp

from typing import Callable

from ..core.simulate import TimeSpan
from ..utils.base import (
    DynamicFunctionStoring,
    ComponentCompositesHandler
)
from ..utils.constants import (
    ALLOWED_BEHAVIORS
)

###############################################


# Describe a desired behavior for a set percentile-based time interval
class Behavior:
    def __init__(self, behaviorType: str, startBound: float = None, endBound: float = None, weight: float = 1.0, resolution: int = 20, requiredMatch: float = 1.0):
        behaviorType = behaviorType.upper()
        if behaviorType in ALLOWED_BEHAVIORS:
            self.BehaviorType = behaviorType
        else:
            raise ValueError(f"'{behaviorType}' is not recognized as a legal behavior type.")
        
        if requiredMatch <= 0.0:
            raise ValueError("Invalid matching point fraction. Behaviors must have a requiredMatch value greater than 0.0 (0%).")
        elif requiredMatch > 1.0:
            raise ValueError("Invalid matching point fraction. Behaviors must have a requiredMatch value less than or equal to 1.0 (100%).")
        
        if startBound is None:
            startBound = 0
        if endBound is None:
            endBound = float('inf')
        if endBound <= startBound:
            raise ValueError("The starting time of your behavior must be less than the ending time of your behavior.")

        self.StartBound = startBound
        self.EndBound = endBound

        self.Weight = weight
        self.Resolution = resolution
        self.PointCount = max(1, int(jnp.round(requiredMatch*resolution)))

        self.EvaluationPoints = jnp.linspace(self.StartBound, self.EndBound, self.Resolution)
    

    def isInBounds(self, timePoint: float):
        return jnp.logical_and(timePoint >= self.StartBound, timePoint <= self.EndBound)


    def __str__(self):
        endBoundLabel = "\u221E" if self.EndBound == float('inf') else f"{self.EndBound:.4g}"
        return f"{self.BehaviorType} Behavior: [{self.StartBound:.4g}, {endBoundLabel})"


# Assign a list of behaviors with a given and a normalization method (by default, it is the fraction of the component's initial concentration)
class BehaviorGroup(ComponentCompositesHandler, DynamicFunctionStoring):
    def __init__(self, compNames: str | list[str], behaviors: list[Behavior], normalizeFn: Callable[[float | int, dict], float], compositesAllowed: bool = False, weight: float = 1.0):
        ComponentCompositesHandler.__init__(self, compNames, compositesAllowed)

        self.Weight = weight
        
        self.normalizeValue = normalizeFn
        
        if len(behaviors) < 1:
            raise ValueError("Must include at least one behavior when optimizing for behaviors.")
        
        self.Behaviors = sorted(behaviors, key = self._getBehaviorOrderPriority)

        for i in range(len(self.Behaviors)-1):
            curBehavior = self.Behaviors[i]
            nextBehavior = self.Behaviors[i+1]
            if curBehavior.EndBound > nextBehavior.StartBound:
                raise ValueError(f"The following behaviors overlap in time and cannot coexist in this behavior group:\n  {curBehavior}\n  {nextBehavior}")
        
        self.EvaluationPoints = jnp.concatenate([behavior.EvaluationPoints for behavior in self.Behaviors])
        self.EvaluationIsHigh = jnp.concatenate([jnp.array([True if behavior.BehaviorType == "HIGH" else False for _ in range(behavior.Resolution)], dtype = bool) for behavior in self.Behaviors])
        
        allBehaviorWeight = sum([b.Weight for b in self.Behaviors])
        self.EvaluationPointWeights = jnp.concatenate([jnp.full(behavior.Resolution, behavior.Weight / allBehaviorWeight / behavior.Resolution) for behavior in self.Behaviors])
        
    def _getBehaviorOrderPriority(self, x):
            return x.StartBound



class BehaviorTimeCourse:
    def __init__(self, timeCourse: list[TimeSpan], behaviorGroups: list[BehaviorGroup], weight = 1.0):
        self.TimeCourse = timeCourse
        self.BehaviorGroups = behaviorGroups
        self.Weight = weight

        self.UniqueEvaluationPoints = jnp.sort(jnp.unique(jnp.concatenate([bG.EvaluationPoints for bG in self.BehaviorGroups]))) # Create a list of every unique evaluation point.
        self.EvaluationPointMappingGroups = [jnp.searchsorted(self.UniqueEvaluationPoints, bG.EvaluationPoints) for bG in self.BehaviorGroups] # Locate all the positions of the final evaluation points for each of the evaluation point groups.
        
        self.BehaviorGroupWeights = jnp.array([bG.Weight for bG in self.BehaviorGroups])
        self.BehaviorGroupWeights /= jnp.sum(self.BehaviorGroupWeights)



################ Initialization ###############
#                                             #
#               By: Colin Yancey              #
#            Created September 2025           #
#            Last Edited: 02/24/2026          #
#                  Schulman Lab               #
#            Johns Hopkins University         #
#                                             #
###############################################

"""
Modules:
    crnConstructor: Core classes for defining reactions, components, and chemical reaction networks.
    wrappers: Domain-specific network constructors.

Example:
    >>> rxn1 = Reaction(5e-5, "A + B -> C")
    >>> rxn2 = Reaction(1e-2, "C + D <-> CD", backwardRate = 1e-0)
    >>> rxn3 = Reaction(3e-2, "CD -> D + E")

    >>> compA = Component("A", 100)
    >>> compB = Component("B", 80)
    >>> compD = Component("D", 5)

    >>> newNetwork = ReactionNetwork()
    >>> newNetwork.addRxns([rxn1, rxn2, rxn3])
    >>> newNetwork.addComponents([compA, compB, compD])
    >>> newNetwork.simulateReactionFn(1000)
"""

from .crnConstructor import (
    Reaction, 
    Component,
    ReactionNetwork,

    # To be used internally
    _getRxnStr
)

__all__ = [
    "Reaction",
    "Component", 
    "ReactionNetwork",

    "_getRxnStr"
]